"""Calculate jet engine emissions using the ICAO Aircraft Emissions Databank (EDB).

Functions without a subscript "_" can be used independently outside .eval()
"""

from __future__ import annotations

import dataclasses
import pathlib
import warnings
from typing import Any, NoReturn, overload

import numpy as np
import numpy.typing as npt
import pandas as pd

from pycontrails.core import flight
from pycontrails.core.flight import Flight
from pycontrails.core.fuel import Fuel, JetA, SAFBlend
from pycontrails.core.interpolation import EmissionsProfileInterpolator
from pycontrails.core.met import MetDataset
from pycontrails.core.met_var import AirTemperature, SpecificHumidity
from pycontrails.core.models import Model, ModelParams
from pycontrails.models.emissions import black_carbon, ffm2
from pycontrails.models.humidity_scaling import HumidityScaling
from pycontrails.physics import constants, jet, units

_path_to_static = pathlib.Path(__file__).parent / "static"


@dataclasses.dataclass
class EmissionsParams(ModelParams):
    """:class:`Emissions` model parameters."""

    #: Default paths
    edb_engine_path: str | pathlib.Path = _path_to_static / "edb-gaseous-v28c-engines.csv"
    edb_nvpm_path: str | pathlib.Path = _path_to_static / "edb-nvpm-v28c-engines.csv"

    #: Default nvpm_ei_n value if calculation fails.
    default_nvpm_ei_n: float = 1e15

    #: Fuel type
    #: This parameter must be set on init even for :class:`FlightModels`
    #: because it is used in assembling the nvPM profiles.
    fuel: Fuel = dataclasses.field(default_factory=JetA)

    #: Humidity scaling
    humidity_scaling: HumidityScaling | None = None


class Emissions(Model):
    """Emissions handling using ICAO Emissions Databank (EDB) and black carbon correlations.

    See :mod:`pycontrails.models.emissions.black_carbon`

    Parameters
    ----------
    met : MetDataset | None, optional
        Met data, by default None.
    params : dict[str, Any] | None, optional
        Model parameters, by default None.
    params_kwargs : Any
        Model parameters passed as keyword arguments.

    References
    ----------
    - :cite:`leeContributionGlobalAviation2021`
    - :cite:`schumannDehydrationEffectsContrails2015`
    - :cite:`stettlerGlobalCivilAviation2013`
    - :cite:`wilkersonAnalysisEmissionData2010`
    """

    name = "emissions"
    long_name = "ICAO Emissions Databank (EDB)"
    met_variables = AirTemperature, SpecificHumidity
    default_params = EmissionsParams

    # For performance reasons, these dictionaries should remain class variables
    # They will only be created once (see logic in __init__)
    edb_engine_gaseous: dict[str, EDBGaseous]
    edb_engine_nvpm: dict[str, EDBnvpm]

    def __init__(
        self,
        met: MetDataset | None = None,
        params: dict[str, Any] | None = None,
        **params_kwargs: Any,
    ) -> None:
        super().__init__(met, params, **params_kwargs)

        # Set class variable with engine parameters if not yet loaded
        if not hasattr(self, "edb_engine_gaseous"):
            type(self).edb_engine_gaseous = get_engine_params_from_edb(
                self.params["edb_engine_path"]
            )

        # Set class variable with nvPM emissions profile if not yet loaded
        if not hasattr(self, "edb_engine_nvpm"):
            type(self).edb_engine_nvpm = get_engine_nvpm_profile_from_edb(
                self.params["edb_nvpm_path"]
            )

    #: Last Flight evaluated in model
    source: Flight

    @overload
    def eval(self, source: Flight, **params: Any) -> Flight:
        ...

    @overload
    def eval(self, source: None = None, **params: Any) -> NoReturn:
        ...

    def eval(self, source: Flight | None = None, **params: Any) -> Flight:
        """Calculate the emissions data for ``source``.

        Parameter ``source`` must contain each of the variables:
            - air_temperature
            - specific_humidity
            - true_airspeed
            - fuel_flow

        In addition, ``source.attrs`` should contain variables
            - engine_uid
            - n_engine

        If 'engine_uid' is not provided in the flight attribute or not available in the ICAO EDB,
        constant emission indices will be assumed for NOx, CO, HC, and nvPM mass and number.

        The computed pollutants include carbon dioxide (CO2), nitrogen oxide (NOx),
        carbon monoxide (CO), hydrocarbons (HC), non-volatile particulate matter
        (nvPM) mass and number, sulphur oxides (SOx), sulphates (S) and organic carbon (OC).

        Parameters
        ----------
        source : Flight
            Flight to evaluate
        **params : Any
            Overwrite model parameters before eval

        Returns
        -------
        Flight
            Flight with attached emissions data

        Raises
        ------
        AttributeError
            Raised if :attr:`fuel` on ``flight`` is incompatible with the model parameter "fuel".
        KeyError
            Raised if ``flight`` already contains "thrust_setting" or "nvpm_ei_n" variables.
        """
        self.update_params(params)
        self.set_source(source)
        self.source = self.require_source_type(Flight)

        # Set air_temperature and specific_humidity if not already set
        scale_humidity = (self.params["humidity_scaling"] is not None) and (
            "specific_humidity" not in self.source
        )
        self.set_source_met()

        # Only enhance humidity if it wasn't already present on source
        if scale_humidity:
            self.params["humidity_scaling"].eval(self.source, copy_source=False)

        # Confirm that fuel types match as sanity check
        # FIXME: Should we use self.get_source_param here? This is consistent with how
        # other models deal with possible parameter ambiguity.
        if self.source.fuel != self.params["fuel"]:
            raise AttributeError(
                f"Fuel attribute on Flight ({self.source.fuel.fuel_name}) "
                "is not the same as Emissions model parameter "
                f"({self.params['fuel'].fuel_name}). The 'fuel' model parameter "
                "must be set on init to assemble the nvPM emissions profiles "
                "for each engine type."
            )

        # Ensure that flight has the required variables defined as attrs or columns
        # We could support calculating true_airspeed in the same way as done by BADAFlight
        # Right now, the pathway to using this model is by chaining with BADAFlight,
        # so we don't need to support it yet.
        self.source.ensure_vars(("true_airspeed", "fuel_flow"))

        engine_uid = self.source.attrs.get("engine_uid")
        if engine_uid is None:
            warnings.warn(
                "No 'engine_uid' found on source attrs. A constant emissions will be used."
            )

        try:
            fuel_flow_per_engine = self.source.get_data_or_attr("fuel_flow_per_engine")
        except KeyError:
            # Try to keep vector and attrs data consistent here
            try:
                ff = self.source["fuel_flow"]
            except KeyError:
                ff = self.source.attrs["fuel_flow"]
                fuel_flow_per_engine = ff / self.source.attrs["n_engine"]
                self.source.attrs["fuel_flow_per_engine"] = fuel_flow_per_engine
            else:
                fuel_flow_per_engine = ff / self.source.attrs["n_engine"]
                self.source["fuel_flow_per_engine"] = fuel_flow_per_engine

        # Attach thrust setting
        if "thrust_setting" not in self.source:
            try:
                edb_gaseous = self.edb_engine_gaseous[engine_uid]  # type: ignore[index]
            except KeyError:
                self.source["thrust_setting"] = np.full(shape=len(self.source), fill_value=np.nan)
            else:
                self.source["thrust_setting"] = get_thrust_setting(
                    edb_gaseous,
                    fuel_flow_per_engine=fuel_flow_per_engine,
                    air_pressure=self.source.air_pressure,
                    air_temperature=self.source["air_temperature"],
                    true_airspeed=self.source.get_data_or_attr("true_airspeed"),
                )

        self._gaseous_emission_indices(engine_uid)
        self._nvpm_emission_indices(engine_uid)
        self._total_pollutant_emissions()
        return self.source

    def _gaseous_emission_indices(self, engine_uid: str | None) -> None:
        """Calculate EI's for nitrogen oxide (NOx), carbon monoxide (CO) and hydrocarbons (HC).

        This method attaches the following variables to the underlying :attr:`flight`:

        - `nox_ei`
        - `co_ei`
        - `hc_ei`

        Parameters
        ----------
        engine_uid : str
            Engine unique identification number from the ICAO EDB
        """
        try:
            edb_gaseous = self.edb_engine_gaseous[engine_uid]  # type: ignore[index]
        except KeyError:
            self._gaseous_emissions_constant()
        else:
            self._gaseous_emissions_ffm2(edb_gaseous)

    def _gaseous_emissions_ffm2(self, edb_gaseous: EDBGaseous) -> None:
        """Calculate gaseous emissions using the FFM2 methodology.

        This method attaches the following variables to the underlying :attr:`flight`:

        - `nox_ei`
        - `co_ei`
        - `hc_ei`

        Parameters
        ----------
        edb_gaseous : EDBGaseous
            EDB gaseous data
        """
        self.source.attrs["gaseous_data_source"] = "FFM2"

        fuel_flow_per_engine = self.source.get_data_or_attr("fuel_flow_per_engine")
        true_airspeed = self.source.get_data_or_attr("true_airspeed")
        air_temperature = self.source["air_temperature"]

        # Emissions indices
        self.source["nox_ei"] = nitrogen_oxide_emissions_index_ffm2(
            edb_gaseous,
            fuel_flow_per_engine,
            true_airspeed,
            self.source.air_pressure,
            air_temperature,
            self.source["specific_humidity"],
        )

        self.source["co_ei"] = carbon_monoxide_emissions_index_ffm2(
            edb_gaseous,
            fuel_flow_per_engine,
            true_airspeed,
            self.source.air_pressure,
            air_temperature,
        )

        self.source["hc_ei"] = hydrocarbon_emissions_index_ffm2(
            edb_gaseous,
            fuel_flow_per_engine,
            true_airspeed,
            self.source.air_pressure,
            air_temperature,
        )

    def _gaseous_emissions_constant(self) -> None:
        """Fill gaseous emissions data with default values.

        This method attaches the following variables to the underlying :attr:`flight`:

        - `nox_ei`
        - `co_ei`
        - `hc_ei`

        Assumes constant emission indices for nitrogen oxide, carbon monoxide and
        hydrocarbon for a given aircraft-engine pair if data is not available in the ICAO EDB.

        - NOx EI = 15.14 g-NOx/kg-fuel (Table 1 of Lee et al., 2020)
        - CO EI = 3.61 g-CO/kg-fuel (Table 1 of Wilkerson et al., 2010),
        - HC EI = 0.520 g-HC/kg-fuel (Table 1 of Wilkerson et al., 2010)

        References
        ----------
        - :cite:`leeContributionGlobalAviation2021`
        - :cite:`wilkersonAnalysisEmissionData2010`
        """
        self.source.attrs["gaseous_data_source"] = "Constant"

        nox_ei = np.full(shape=len(self.source), fill_value=15.14)
        co_ei = np.full(shape=len(self.source), fill_value=3.61)
        hc_ei = np.full(shape=len(self.source), fill_value=0.520)

        self.source["nox_ei"] = nox_ei * 1e-3  # g-NOx/kg-fuel to kg-NOx/kg-fuel
        self.source["co_ei"] = co_ei * 1e-3  # g-CO/kg-fuel to kg-CO/kg-fuel
        self.source["hc_ei"] = hc_ei * 1e-3  # g-HC/kg-fuel to kg-HC/kg-fuel

    def _nvpm_emission_indices(self, engine_uid: str | None) -> None:
        """Calculate emission indices for nvPM mass and number.

        This method attaches the following variables to the underlying :attr:`source`.
            - nvpm_ei_m
            - nvpm_ei_n

        In addition, ``nvpm_data_source`` is attached to the ``source.attrs``.

        Parameters
        ----------
        engine_uid : str
            Engine unique identification number from the ICAO EDB
        """
        if "nvpm_ei_n" in self.source and "nvpm_ei_m" in self.source:
            return  # early exit if values already exist

        if (edb_nvpm := self.edb_engine_nvpm.get(engine_uid)) is not None:  # type: ignore[arg-type]
            nvpm_data_source, nvpm_ei_m, nvpm_ei_n = self._nvpm_emission_indices_edb(edb_nvpm)
        elif (edb_gaseous := self.edb_engine_gaseous.get(engine_uid)) is not None:  # type: ignore[arg-type]  # noqa: E501
            nvpm_data_source, nvpm_ei_m, nvpm_ei_n = self._nvpm_emission_indices_sac(edb_gaseous)
        else:
            if engine_uid is not None:
                warnings.warn(
                    f"Cannot find 'engine_uid' {engine_uid} in EDB. "
                    "A constant emissions will be used."
                )
            nvpm_data_source, nvpm_ei_m, nvpm_ei_n = self._nvpm_emission_indices_constant()

        # Adjust nvPM emission indices if SAF is used.
        if isinstance(self.source.fuel, SAFBlend) and self.source.fuel.pct_blend:
            pct_eim_reduction = black_carbon.nvpm_mass_ei_pct_reduction_due_to_saf(
                self.source.fuel.hydrogen_content, self.source["thrust_setting"]
            )
            pct_ein_reduction = black_carbon.nvpm_number_ei_pct_reduction_due_to_saf(
                self.source.fuel.hydrogen_content, self.source["thrust_setting"]
            )

            nvpm_ei_m *= 1.0 + pct_eim_reduction / 100.0
            nvpm_ei_n *= 1.0 + pct_ein_reduction / 100.0

        self.source.attrs["nvpm_data_source"] = nvpm_data_source
        self.source.setdefault("nvpm_ei_m", nvpm_ei_m)
        self.source.setdefault("nvpm_ei_n", nvpm_ei_n)

    def _nvpm_emission_indices_edb(
        self, edb_nvpm: EDBnvpm
    ) -> tuple[str, npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        """Calculate emission indices for nvPM mass and number.

        This method uses data from the ICAO EDB along with the T4/T2 methodology.

        Parameters
        ----------
        edb_nvpm : EDBnvpm
            EDB nvPM data.

        Returns
        -------
        nvpm_data_source : str
            Source of nvpm data.
        nvpm_ei_m : npt.NDArray[np.float_]
            Non-volatile particulate matter (nvPM) mass emissions index, [:math:`kg/kg_{fuel}`]
        nvpm_ei_n : npt.NDArray[np.float_]
            Black carbon number emissions index, [:math:`kg_{fuel}^{-1}`]

        References
        ----------
        - :cite:`teohTargetedUseSustainable2022`
        """
        nvpm_data_source = "ICAO EDB"

        # Emissions indices
        return nvpm_data_source, *get_nvpm_emissions_index_edb(
            edb_nvpm,
            true_airspeed=self.source.get_data_or_attr("true_airspeed"),
            air_temperature=self.source["air_temperature"],
            air_pressure=self.source.air_pressure,
            thrust_setting=self.source["thrust_setting"],
            q_fuel=self.source.fuel.q_fuel,
        )

    def _nvpm_emission_indices_sac(
        self, edb_gaseous: EDBGaseous
    ) -> tuple[str, npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        """Calculate EIs for nvPM mass and number assuming the profile of single annular combustors.

        nvPM EI_m is calculated using the FOX and ImFOX methods, while the nvPM EI_n
        is calculated using the Fractal Aggregates (FA) model.

        Parameters
        ----------
        edb_gaseous : EDBGaseous
            EDB gaseous data

        Returns
        -------
        nvpm_data_source : str
            Source of nvpm data.
        nvpm_ei_m : npt.NDArray[np.float_]
            Non-volatile particulate matter (nvPM) mass emissions index, [:math:`kg/kg_{fuel}`]
        nvpm_ei_n : npt.NDArray[np.float_]
            Black carbon number emissions index, [:math:`kg_{fuel}^{-1}`]

        References
        ----------
        - :cite:`stettlerGlobalCivilAviation2013`
        - :cite:`abrahamsonPredictiveModelDevelopment2016`
        - :cite:`teohTargetedUseSustainable2022`
        """
        nvpm_data_source = "FA Model"

        # calculate properties
        thrust_setting = self.source["thrust_setting"]
        fuel_flow_per_engine = self.source.get_data_or_attr("fuel_flow_per_engine")
        true_airspeed = self.source.get_data_or_attr("true_airspeed")
        air_temperature = self.source["air_temperature"]

        # Emissions indices
        nvpm_ei_m = nvpm_mass_emissions_index_sac(
            edb_gaseous,
            air_pressure=self.source.air_pressure,
            true_airspeed=true_airspeed,
            air_temperature=air_temperature,
            thrust_setting=thrust_setting,
            fuel_flow_per_engine=fuel_flow_per_engine,
            hydrogen_content=self.source.fuel.hydrogen_content,
        )
        nvpm_gmd = nvpm_geometric_mean_diameter_sac(
            edb_gaseous,
            air_pressure=self.source.air_pressure,
            true_airspeed=true_airspeed,
            air_temperature=air_temperature,
            thrust_setting=thrust_setting,
            q_fuel=self.source.fuel.q_fuel,
        )
        nvpm_ei_n = black_carbon.number_emissions_index_fractal_aggregates(nvpm_ei_m, nvpm_gmd)
        return nvpm_data_source, nvpm_ei_m, nvpm_ei_n

    def _nvpm_emission_indices_constant(
        self,
    ) -> tuple[str, npt.NDArray[np.float_], npt.NDArray[np.float_]]:
        """
        Assume constant emission indices for nvPM mass and number.

        (nvpm_ei_n = 1e15 /kg-fuel) for a given aircraft-engine pair if data
        is not available in the ICAO EDB.

        - nvpm_ei_m = 0.088 g-nvPM/kg-fuel (Table 2 of Stettler et al., 2013)
        - nvpm_ei_n = 1e15 /kg-fuel (Schumann et al., 2015)

        Returns
        -------
        nvpm_data_source : str
            Source of nvpm data.
        nvpm_ei_m : npt.NDArray[np.float_]
            Non-volatile particulate matter (nvPM) mass emissions index, [:math:`kg/kg_{fuel}`]
        nvpm_ei_n : npt.NDArray[np.float_]
            Black carbon number emissions index, [:math:`kg_{fuel}^{-1}`]

        References
        ----------
        - :cite:`stettlerGlobalCivilAviation2013`
        - :cite:`wilkersonAnalysisEmissionData2010`
        - :cite:`schumannDehydrationEffectsContrails2015`
        """
        nvpm_data_source = "Constant"
        nvpm_ei_m = np.full(shape=len(self.source), fill_value=(0.088 * 1e-3))  # g to kg
        nvpm_ei_n = np.full(shape=len(self.source), fill_value=self.params["default_nvpm_ei_n"])
        return nvpm_data_source, nvpm_ei_m, nvpm_ei_n

    def _total_pollutant_emissions(self) -> None:
        # Required variables
        # FIXME: These two variables are already calculated in BADA models
        dt_sec = flight.segment_duration(self.source["time"], dtype=self.source.altitude_ft.dtype)
        fuel_burn = jet.fuel_burn(self.source.get_data_or_attr("fuel_flow"), dt_sec)

        # TODO: these currently overwrite values and will throw warnings

        # Total emissions for each waypoint
        self.source["co2"] = fuel_burn * self.source.fuel.ei_co2
        self.source["h2o"] = fuel_burn * self.source.fuel.ei_h2o
        self.source["so2"] = fuel_burn * self.source.fuel.ei_so2
        self.source["sulphates"] = fuel_burn * self.source.fuel.ei_sulphates
        self.source["oc"] = fuel_burn * self.source.fuel.ei_oc
        self.source["nox"] = fuel_burn * self.source["nox_ei"]
        self.source["co"] = fuel_burn * self.source["co_ei"]
        self.source["hc"] = fuel_burn * self.source["hc_ei"]
        self.source["nvpm_mass"] = fuel_burn * self.source["nvpm_ei_m"]
        self.source["nvpm_number"] = fuel_burn * self.source["nvpm_ei_n"]

        # Total emissions for the flight
        self.source.attrs["total_co2"] = np.nansum(self.source["co2"])
        self.source.attrs["total_h2o"] = np.nansum(self.source["h2o"])
        self.source.attrs["total_so2"] = np.nansum(self.source["so2"])
        self.source.attrs["total_sulphates"] = np.nansum(self.source["sulphates"])
        self.source.attrs["total_oc"] = np.nansum(self.source["oc"])
        self.source.attrs["total_nox"] = np.nansum(self.source["nox"])
        self.source.attrs["total_co"] = np.nansum(self.source["co"])
        self.source.attrs["total_hc"] = np.nansum(self.source["hc"])
        self.source.attrs["total_nvpm_mass"] = np.nansum(self.source["nvpm_mass"])
        self.source.attrs["total_nvpm_number"] = np.nansum(self.source["nvpm_number"])

    def _check_edb_gaseous_availability(
        self,
        engine_uid: str,
        raise_error: bool = True,
    ) -> bool:
        """
        Check if the provided engine is available in the gaseous ICAO EDB.

        Setting ``raise_error`` to True allows functions in this class to be
        used independently outside of :meth:`eval`.

        Parameters
        ----------
        engine_uid: str
            Engine unique identification number from the ICAO EDB
        raise_error: bool
            Raise a KeyError if engine type is not available.

        Returns
        -------
        bool
            True if engine type is available in the gaseous ICAO EDB.

        Raises
        ------
        KeyError
            If engine type is not available in the gaseous ICAO EDB.
        """
        if engine_uid not in self.edb_engine_gaseous:
            if raise_error:
                raise KeyError(
                    f"Engine ({engine_uid}) is not available in the ICAO EDB gaseous database"
                )
            return False
        return True

    def _check_edb_nvpm_availability(
        self,
        engine_uid: str,
        raise_error: bool = True,
    ) -> bool:
        """
        Check if the provided engine is available in the nvPM ICAO EDB.

        Setting ``raise_error`` to True allows functions in this class to be
        used independently outside of :meth:`eval`.

        Parameters
        ----------
        engine_uid: str
            Engine unique identification number from the ICAO EDB
        raise_error: bool
            Raise a KeyError if engine type is not available.

        Returns
        -------
        bool
            True if engine type is available in the nvPM ICAO EDB.

        Raises
        ------
        KeyError
            If engine type is not available in the nvPM ICAO EDB.
        """
        if engine_uid not in self.edb_engine_nvpm:
            if raise_error:
                raise KeyError(
                    f"Engine ({engine_uid}) is not available in the ICAO EDB nvPM database"
                )
            return False
        return True


def nitrogen_oxide_emissions_index_ffm2(
    edb_gaseous: EDBGaseous,
    fuel_flow_per_engine: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    specific_humidity: None | npt.NDArray[np.float_] = None,
) -> npt.NDArray[np.float_]:
    """
    Estimate the nitrogen oxide (NOx) emissions index (EI) using the Fuel Flow Method 2 (FFM2).

    Parameters
    ----------
    edb_gaseous : EDBGaseous
        EDB gaseous data
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    air_pressure : npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    specific_humidity: npt.NDArray[np.float_]
        specific humidity for each waypoint, [:math:`kg_{H_{2}O}/kg_{air}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Nitrogen oxide emissions index for each waypoint, [:math:`kg_{NO_{X}}/kg_{fuel}`]
    """
    res_nox = ffm2.estimate_nox(
        edb_gaseous.log_ei_nox_profile,
        fuel_flow_per_engine,
        true_airspeed,
        air_pressure,
        air_temperature,
        specific_humidity,
    )
    return res_nox * 1e-3  # g-NOx/kg-fuel to kg-NOx/kg-fuel


def carbon_monoxide_emissions_index_ffm2(
    edb_gaseous: EDBGaseous,
    fuel_flow_per_engine: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Estimate the carbon monoxide (CO) emissions index (EI) using the Fuel Flow Method 2 (FFM2).

    Parameters
    ----------
    edb_gaseous : EDBGaseous
        EDB gaseous data
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    air_pressure : npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]

    Returns
    -------
    npt.NDArray[np.float_]
        Carbon monoxide emissions index for each waypoint, [:math:`kg_{CO}/kg_{fuel}`]
    """
    res_co = ffm2.estimate_ei(
        edb_gaseous.log_ei_co_profile,
        fuel_flow_per_engine,
        true_airspeed,
        air_pressure,
        air_temperature,
    )
    return res_co * 1e-3  # g-CO/kg-fuel to kg-CO/kg-fuel


def hydrocarbon_emissions_index_ffm2(
    edb_gaseous: EDBGaseous,
    fuel_flow_per_engine: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Estimate the hydrocarbon (HC) emissions index (EI) using the Fuel Flow Method 2 (FFM2).

    Parameters
    ----------
    edb_gaseous : EDBGaseous
        EDB gaseous data
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    air_pressure : npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature : npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]

    Returns
    -------
    npt.NDArray[np.float_]
        Hydrocarbon emissions index for each waypoint, [:math:`kg_{HC}/kg_{fuel}`]
    """
    res_hc = ffm2.estimate_ei(
        edb_gaseous.log_ei_hc_profile,
        fuel_flow_per_engine,
        true_airspeed,
        air_pressure,
        air_temperature,
    )
    return res_hc * 1e-3  # g-HC/kg-fuel to kg-HC/kg-fuel


def get_nvpm_emissions_index_edb(
    edb_nvpm: EDBnvpm,
    true_airspeed: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
    q_fuel: float,
) -> tuple[npt.NDArray[np.float_], npt.NDArray[np.float_]]:
    r"""Calculate nvPM mass emissions index (nvpm_ei_m) and number emissions index (nvpm_ei_n).

    Interpolate the non-volatile particulate matter (nvPM) mass and number emissions index from
    the emissions profile of a given engine type that is provided by the ICAO EDB.

    The non-dimensional thrust setting (t4_t2) is clipped to the minimum and maximum t4_t2 values
    that is estimated from the four ICAO EDB datapoints to prevent extrapolating the nvPM values.

    Parameters
    ----------
    edb_nvpm : EDBnvpm
        EDB nvPM data
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    air_temperature: npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    air_pressure: npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    thrust_setting : npt.NDArray[np.float_]
        thrust setting
    q_fuel : float
        Lower calorific value (LCV) of fuel, [:math:`J \ kg_{fuel}^{-1}`].

    Returns
    -------
    nvpm_ei_m : npt.NDArray[np.float_]
        Non-volatile particulate matter (nvPM) mass emissions index, [:math:`kg/kg_{fuel}`]
    nvpm_ei_n : npt.NDArray[np.float_]
        Black carbon number emissions index, [:math:`kg_{fuel}^{-1}`]
    """
    # Non-dimensionalized thrust setting
    t4_t2 = jet.thrust_setting_nd(
        true_airspeed,
        thrust_setting,
        air_temperature,
        air_pressure,
        edb_nvpm.pressure_ratio,
        q_fuel,
        cruise=True,
    )

    # Interpolate nvPM EI_m and EI_n
    nvpm_ei_m = edb_nvpm.nvpm_ei_m.interp(t4_t2)
    nvpm_ei_m = nvpm_ei_m * 1e-6  # mg-nvPM/kg-fuel to kg-nvPM/kg-fuel
    nvpm_ei_n = edb_nvpm.nvpm_ei_n.interp(t4_t2)
    return nvpm_ei_m, nvpm_ei_n


def nvpm_mass_emissions_index_sac(
    edb_gaseous: EDBGaseous,
    air_pressure: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
    fuel_flow_per_engine: npt.NDArray[np.float_],
    hydrogen_content: float,
) -> npt.NDArray[np.float_]:
    """Estimate nvPM mass emission index for singular annular combustor (SAC) engines.

    Here, SAC should not be confused with the Schmidt-Appleman Criterion.

    The nvpm_ei_m for SAC is estimated as the mean between a lower bound (80% of the FOX-estimated
    nvpm_ei_m) and an upper bound (150% of ImFOX-estimated nvpm_ei_m).

    Parameters
    ----------
    edb_gaseous : EDBGaseous
        EDB gaseous data
    air_pressure: npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    air_temperature: npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    thrust_setting : npt.NDArray[np.float_]
        thrust setting
    fuel_flow_per_engine: npt.NDArray[np.float_]
        fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    hydrogen_content : float
        Engine unique identification number from the ICAO EDB

    Returns
    -------
    npt.NDArray[np.float_]
        nvPM mass emissions index, [:math:`kg/kg_{fuel}`]
    """
    nvpm_ei_m_fox = black_carbon.mass_emissions_index_fox(
        air_pressure,
        air_temperature,
        true_airspeed,
        fuel_flow_per_engine,
        thrust_setting,
        edb_gaseous.pressure_ratio,
    )
    nvpm_ei_m_imfox = black_carbon.mass_emissions_index_imfox(
        fuel_flow_per_engine, thrust_setting, hydrogen_content
    )
    nvpm_ei_m = 0.5 * (0.8 * nvpm_ei_m_fox + 1.5 * nvpm_ei_m_imfox)
    return nvpm_ei_m * 1e-6  # mg-nvPM/kg-fuel to kg-nvPM/kg-fuel


def nvpm_geometric_mean_diameter_sac(
    edb_gaseous: EDBGaseous,
    air_pressure: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    thrust_setting: npt.NDArray[np.float_],
    q_fuel: float,
) -> npt.NDArray[np.float_]:
    r"""
    Estimate nvPM geometric mean diameter for singular annular combustor (SAC) engines.

    Parameters
    ----------
    edb_gaseous : EDBGaseous
        EDB gaseous data
    air_pressure: npt.NDArray[np.float_]
        pressure altitude at each waypoint, [:math:`Pa`]
    true_airspeed: npt.NDArray[np.float_]
        true airspeed for each waypoint, [:math:`m s^{-1}`]
    air_temperature: npt.NDArray[np.float_]
        ambient temperature for each waypoint, [:math:`K`]
    thrust_setting : npt.NDArray[np.float_]
        thrust setting
    q_fuel : float
        Lower calorific value (LCV) of fuel, [:math:`J \ kg_{fuel}^{-1}`].

    Returns
    -------
    npt.NDArray[np.float_]
        nvPM geometric mean diameter, [:math:`m`]
    """
    nvpm_gmd = black_carbon.geometric_mean_diameter_sac(
        air_pressure,
        air_temperature,
        true_airspeed,
        thrust_setting,
        edb_gaseous.pressure_ratio,
        q_fuel,
        cruise=True,
    )
    return nvpm_gmd * 1e-9  # nm to m


def get_thrust_setting(
    edb_gaseous: EDBGaseous,
    fuel_flow_per_engine: npt.NDArray[np.float_],
    air_pressure: npt.NDArray[np.float_],
    air_temperature: npt.NDArray[np.float_],
    true_airspeed: npt.NDArray[np.float_],
) -> npt.NDArray[np.float_]:
    """
    Approximate the engine thrust setting at cruise conditions.

    The thrust setting is approximated by dividing the fuel mass flow rate
    by the maximum fuel mass flow rate, and clipped to 3% (0.03) and 100% (1)
    respectively to account for unrealistic values.

    Parameters
    ----------
    edb_gaseous : EDBGaseous
        EDB gaseous data
    fuel_flow_per_engine: npt.NDArray[np.float_]
        Fuel mass flow rate per engine, [:math:`kg s^{-1}`]
    air_pressure: npt.NDArray[np.float_]
        Pressure altitude at each waypoint, [:math:`Pa`]
    air_temperature: npt.NDArray[np.float_]
        Ambient temperature for each waypoint, [:math:`K`]
    true_airspeed: npt.NDArray[np.float_]
        True airspeed for each waypoint, [:math:`m s^{-1}`]

    Returns
    -------
    npt.NDArray[np.float_]
        Engine thrust setting. Returns ``np.nan`` if engine data is
        not available in the ICAO EDB dataset.
    """
    theta_amb = jet.temperature_ratio(air_temperature)
    delta_amb = jet.pressure_ratio(air_pressure)
    mach_num = units.tas_to_mach_number(true_airspeed, air_temperature)
    fuel_flow_per_engine = jet.equivalent_fuel_flow_rate_at_sea_level(
        fuel_flow_per_engine, theta_amb, delta_amb, mach_num
    )

    thrust_setting = fuel_flow_per_engine / edb_gaseous.ff_100
    thrust_setting.clip(0.03, 1, out=thrust_setting)  # clip in place
    return thrust_setting


class EDBGaseous:
    """Gaseous emissions data.

    -------------------------------------
    ENGINE IDENTIFICATION AND TYPE:
    -------------------------------------
    manufacturer: str
        engine manufacturer
    engine_name: str
        name of engine
    combustor: str
        description of engine combustor

    -------------------------------------
    ENGINE CHARACTERISTICS:
    -------------------------------------
    bypass_ratio: float
        engine bypass ratio
    pressure_ratio: float
        engine pressure ratio
    rated_thrust: float
        rated thrust of engine, [:math:`kN`]

    -------------------------------------
    FUEL CONSUMPTION:
    -------------------------------------
    ff_7: float
        fuel mass flow rate at 7% thrust setting, [:math:`kg s^{-1}`]
    ff_30: float
        fuel mass flow rate at 30% thrust setting, [:math:`kg s^{-1}`]
    ff_85: float
        fuel mass flow rate at 85% thrust setting, [:math:`kg s^{-1}`]
    ff_100: float
        fuel mass flow rate at 100% thrust setting, [:math:`kg s^{-1}`]

    -------------------------------------
    EMISSIONS:
    -------------------------------------
    ei_nox_7: float
        NOx emissions index at 7% thrust setting, [:math:`g_{NO_{X}}/kg_{fuel}`]
    ei_nox_30: float
        NOx emissions index at 30% thrust setting, [:math:`g_{NO_{X}}/kg_{fuel}`]
    ei_nox_85: float
        NOx emissions index at 85% thrust setting, [:math:`g_{NO_{X}}/kg_{fuel}`]
    ei_nox_100: float
        NOx emissions index at 100% thrust setting, [:math:`g_{NO_{X}}/kg_{fuel}`]

    ei_co_7: float
        CO emissions index at 7% thrust setting, [:math:`g_{CO}/kg_{fuel}`]
    ei_co_30: float
        CO emissions index at 30% thrust setting, [:math:`g_{CO}/kg_{fuel}`]
    ei_co_85: float
        CO emissions index at 85% thrust setting, [:math:`g_{CO}/kg_{fuel}`]
    ei_co_100: float
        CO emissions index at 100% thrust setting, [:math:`g_{CO}/kg_{fuel}`]

    ei_hc_7: float
        HC emissions index at 7% thrust setting, [:math:`g_{HC}/kg_{fuel}`]
    ei_hc_30: float
        HC emissions index at 30% thrust setting, [:math:`g_{HC}/kg_{fuel}`]
    ei_hc_85: float
        HC emissions index at 85% thrust setting, [:math:`g_{HC}/kg_{fuel}`]
    ei_hc_100: float
        HC emissions index at 100% thrust setting, [:math:`g_{HC}/kg_{fuel}`]

    sn_7: float
        smoke number at 7% thrust setting
    sn_30: float
        smoke number at 30% thrust setting
    sn_85: float
        smoke number at 85% thrust setting
    sn_100: float
        smoke number at 100% thrust setting
    sn_max: float
        maximum smoke number value across the range of thrust setting
    """

    def __init__(self, df_engine: pd.Series) -> None:
        # Engine identification and type
        self.manufacturer: str = df_engine["Manufacturer"]
        self.engine_name: str = df_engine["Engine Identification"]
        self.combustor: str = df_engine["Combustor Description"]

        # Engine characteristics
        self.bypass_ratio: float = df_engine["B/P Ratio"]
        self.pressure_ratio: float = df_engine["Pressure Ratio"]
        self.rated_thrust: float = df_engine["Rated Thrust (kN)"]

        # Fuel consumption
        self.ff_7: float = df_engine["Fuel Flow Idle (kg/sec)"]
        self.ff_30: float = df_engine["Fuel Flow App (kg/sec)"]
        self.ff_85: float = df_engine["Fuel Flow C/O (kg/sec)"]
        self.ff_100: float = df_engine["Fuel Flow T/O (kg/sec)"]

        # Emissions
        self.ei_nox_7: float = df_engine["NOx EI Idle (g/kg)"]
        self.ei_nox_30: float = df_engine["NOx EI App (g/kg)"]
        self.ei_nox_85: float = df_engine["NOx EI C/O (g/kg)"]
        self.ei_nox_100: float = df_engine["NOx EI T/O (g/kg)"]

        self.ei_co_7: float = df_engine["CO EI Idle (g/kg)"]
        self.ei_co_30: float = df_engine["CO EI App (g/kg)"]
        self.ei_co_85: float = df_engine["CO EI C/O (g/kg)"]
        self.ei_co_100: float = df_engine["CO EI T/O (g/kg)"]

        self.ei_hc_7: float = df_engine["HC EI Idle (g/kg)"]
        self.ei_hc_30: float = df_engine["HC EI App (g/kg)"]
        self.ei_hc_85: float = df_engine["HC EI C/O (g/kg)"]
        self.ei_hc_100: float = df_engine["HC EI T/O (g/kg)"]

        self.sn_7: float = df_engine["SN Idle"]
        self.sn_30: float = df_engine["SN App"]
        self.sn_85: float = df_engine["SN C/O"]
        self.sn_100: float = df_engine["SN T/O"]
        self.sn_max: float = df_engine["SN Max"]

        # Emissions profile
        self.log_ei_nox_profile = ffm2.nitrogen_oxide_emissions_index_profile(
            self.ff_7,
            self.ff_30,
            self.ff_85,
            self.ff_100,
            self.ei_nox_7,
            self.ei_nox_30,
            self.ei_nox_85,
            self.ei_nox_100,
        )
        self.log_ei_co_profile = ffm2.co_hc_emissions_index_profile(
            self.ff_7,
            self.ff_30,
            self.ff_85,
            self.ff_100,
            self.ei_co_7,
            self.ei_co_30,
            self.ei_co_85,
            self.ei_co_100,
        )

        self.log_ei_hc_profile = ffm2.co_hc_emissions_index_profile(
            self.ff_7,
            self.ff_30,
            self.ff_85,
            self.ff_100,
            self.ei_hc_7,
            self.ei_hc_30,
            self.ei_hc_85,
            self.ei_hc_100,
        )


class EDBnvpm:
    """A data class for EDB nvPM data.

    -------------------------------------
    ENGINE IDENTIFICATION AND TYPE:
    -------------------------------------
    manufacturer: str
        engine manufacturer
    engine_name: str
        name of engine
    combustor: str
        description of engine combustor

    -------------------------------------
    ENGINE CHARACTERISTICS:
    -------------------------------------
    pressure_ratio: float
        engine pressure ratio

    -------------------------------------
    nvPM EMISSIONS:
    -------------------------------------
    nvpm_ei_m: EmissionsProfileInterpolator
         non-volatile PM mass emissions index profile (mg/kg) vs.
         non-dimensionalized thrust setting (t4_t2)
    nvpm_ei_n: EmissionsProfileInterpolator
        non-volatile PM number emissions index profile (1/kg) vs.
        non-dimensionalized thrust setting (t4_t2)
    """

    def __init__(self, df_engine: pd.Series) -> None:
        # Engine identification and type
        self.manufacturer: str = df_engine["Manufacturer"]
        self.engine_name: str = df_engine["Engine Identification"]
        self.combustor: str = df_engine["Combustor Description"]

        # Engine characteristics
        self.pressure_ratio: float = df_engine["Pressure Ratio"]

        # nvPM Emissions
        self.nvpm_ei_m, self.nvpm_ei_n = self._get_nvpm_emissions_profile(df_engine)

    def _get_nvpm_emissions_profile(
        self, df_engine: pd.Series
    ) -> tuple[EmissionsProfileInterpolator, EmissionsProfileInterpolator]:
        # Extract fuel flow
        cols = [
            "Fuel Flow Idle (kg/sec)",
            "Fuel Flow App (kg/sec)",
            "Fuel Flow C/O (kg/sec)",
            "Fuel Flow T/O (kg/sec)",
        ]
        fuel_flow = df_engine[cols].to_numpy(dtype=float)
        fuel_flow_max = fuel_flow[-1]

        # Extract nvPM emissions arrays
        cols = [
            "nvPM EImass_SL Idle (mg/kg)",
            "nvPM EImass_SL App (mg/kg)",
            "nvPM EImass_SL C/O (mg/kg)",
            "nvPM EImass_SL T/O (mg/kg)",
        ]
        nvpm_ei_m = df_engine[cols].to_numpy(dtype=float)
        cols = [
            "nvPM EInum_SL Idle (#/kg)",
            "nvPM EInum_SL App (#/kg)",
            "nvPM EInum_SL C/O (#/kg)",
            "nvPM EInum_SL T/O (#/kg)",
        ]
        nvpm_ei_n = df_engine[cols].to_numpy(dtype=float)

        # Modify nvPM mass and number emissions profile to account for lean-burn stage
        is_staged_combustor = df_engine["Combustor Description"] in ["DAC", "TAPS", "TAPS II"]

        if is_staged_combustor:
            # In this case, all of our interpolators will have size 5
            fuel_flow = np.insert(fuel_flow, 2, (fuel_flow[1] * 1.001))
            nvpm_ei_n_lean_burn = np.mean(nvpm_ei_n[2:])
            nvpm_ei_n = np.r_[nvpm_ei_n[:2], [nvpm_ei_n_lean_burn] * 3]
            nvpm_ei_m_lean_burn = np.mean(nvpm_ei_m[2:])
            nvpm_ei_m = np.r_[nvpm_ei_m[:2], [nvpm_ei_m_lean_burn] * 3]

        thrust_setting = fuel_flow / fuel_flow_max
        avg_temp = (df_engine["Ambient Temp Min (K)"] + df_engine["Ambient Temp Max (K)"]) / 2

        t4_t2 = jet.thrust_setting_nd(
            true_airspeed=0.0,
            thrust_setting=thrust_setting,
            T=avg_temp,
            p=constants.p_surface,
            pressure_ratio=self.pressure_ratio,
            q_fuel=df_engine["Fuel Heat of Combustion (MJ/kg)"] * 1e6,
            cruise=False,
        )

        # nvPM mass emissions profile
        nvpm_ei_m_interp = EmissionsProfileInterpolator(t4_t2, nvpm_ei_m)
        nvpm_ei_n_interp = EmissionsProfileInterpolator(t4_t2, nvpm_ei_n)
        return nvpm_ei_m_interp, nvpm_ei_n_interp


def get_engine_params_from_edb(file_path_edb_processed: str) -> dict[str, EDBGaseous]:
    """Read EDB file into a dictionary of the form ``{engine_uid: gaseous_data}``.

    Parameters
    ----------
    file_path_edb_processed : str
        Path to EDB csv.

    Returns
    -------
    dict[str, EDBGaseous]
        Mapping from aircraft type to gaseous emissions data for engine.
    """
    df = load_edb_dataset(file_path_edb_processed)
    return {engine_uid: EDBGaseous(df_engine) for engine_uid, df_engine in df.iterrows()}


def get_engine_nvpm_profile_from_edb(file_edb_nvpm_path: str) -> dict[str, EDBnvpm]:
    """Read EDB file into a dictionary of the form ``{engine_uid: npvm_data}``.

    Parameters
    ----------
    file_edb_nvpm_path : str
        Path to EDB csv.
    fuel : Fuel
        Fuel instance for engines.

    Returns
    -------
    dict[str, EDBnvpm]
        Mapping from aircraft type to nvPM data for engine.
    """
    df = load_edb_dataset(file_edb_nvpm_path)
    return {engine_uid: EDBnvpm(df_engine) for engine_uid, df_engine in df.iterrows()}


def load_edb_dataset(file_name: str) -> pd.DataFrame:
    """Read EDB file to DataFrame.

    Parameters
    ----------
    file_name : str
        Path to EDB csv.

    Returns
    -------
    pd.DataFrame
        EDB data with index column "UID No".
    """
    return pd.read_csv(file_name, index_col=0)



"""Support for humidity scaling methodologies on ERA5 data."""

from __future__ import annotations

import abc
import dataclasses
import warnings
from typing import Any, NoReturn, overload

import numpy as np
import xarray as xr
from overrides import overrides

from pycontrails.core.met import MetDataset
from pycontrails.core.models import Model, ModelParams
from pycontrails.core.vector import GeoVectorDataset
from pycontrails.physics import constants, thermo, units
from pycontrails.utils.types import ArrayLike


def _rhi_over_q(air_temperature: ArrayLike, air_pressure: ArrayLike) -> ArrayLike:
    """Compute the quotient RHi / q."""
    return air_pressure * (constants.R_v / constants.R_d) / thermo.e_sat_ice(air_temperature)


class HumidityScaling(Model):
    """Support for standardizing humidity scaling methodologies.

    The method :meth:`scale` or :meth:`eval` should be called immediately
    after interpolation over meteorology data.

    .. versionadded:: 0.27.0
    """

    #: Variables required in addition to specific_humidity, air_temperature, and air_pressure
    #: These are either :class:`ModelParams` specific to scaling, or variables that should
    #: be extracted from :meth:`eval` parameter ``source``.
    scaler_specific_keys: tuple[str, ...] = tuple()

    @property
    @abc.abstractmethod
    def formula(self) -> str:
        """Serializable formula for humidity scaler."""

    @property
    def description(self) -> dict[str, Any]:
        """Get description for instance."""
        params = {k: v for k in self.scaler_specific_keys if (v := self.params.get(k)) is not None}
        return {"name": self.name, "formula": self.formula, **params}

    # Used by pycontrails.utils.json.NumpyEncoder
    # This ensures model parameters are serializable to JSON
    to_json = description

    @abc.abstractmethod
    def scale(
        self,
        specific_humidity: ArrayLike,
        air_temperature: ArrayLike,
        air_pressure: ArrayLike,
        **kwargs: ArrayLike,
    ) -> tuple[ArrayLike, ArrayLike]:
        r"""Compute scaled specific humidity and RHi.

        See docstring for the implementing subclass for specific methodology.

        Parameters
        ----------
        specific_humidity : ArrayLike
            Unscaled specific relative humidity, [:math:`kg \ kg^{-1}`]. Typically,
            this is interpolated meteorology data.
        air_temperature : ArrayLike
            Air temperature, [:math:`K`]. Typically, this is interpolated meteorology
            data.
        air_pressure : ArrayLike
            Pressure, [:math:`Pa`]
        kwargs : ArrayLike
            Other keyword-only variables and model parameters used by the formula.

        Returns
        -------
        specific_humidity : ArrayLike
            Scaled specific humidity.
        rhi : ArrayLike
            Scaled relative humidity over ice.

        See Also
        --------
        :meth:`eval`
        """

    @overload
    def eval(self, source: GeoVectorDataset, **params: Any) -> GeoVectorDataset:
        ...

    @overload
    def eval(self, source: MetDataset, **params: Any) -> MetDataset:
        ...

    @overload
    def eval(self, source: None = ..., **params: Any) -> NoReturn:
        ...

    def eval(
        self, source: GeoVectorDataset | MetDataset | None = None, **params: Any
    ) -> GeoVectorDataset | MetDataset:
        """Scale specific humidity values on ``source``.

        This method mutates the parameter ``source`` by modifying its
        "specific_humidity" variable and by attaching an "rhi" variable. Set
        model parameter ``copy_source=True`` to avoid mutating ``source``.

        Parameters
        ----------
        source : GeoVectorDataset | MetDataset
            Data with variables "specific_humidity", "air_temperature",
            and any variables defined by :attr:`scaler_specific_keys`.
        **params : Any
            Overwrite model parameters before eval

        Returns
        -------
        GeoVectorDataset | MetDataset
            Source data with updated "specific_humidity" and "rhi". If ``source``
            is :class:`GeoVectorDataset`, "air_pressure" data is also attached.

        See Also
        --------
        :meth:`scale`
        """
        self.update_params(params)
        if source is None:
            raise TypeError("Parameter source must be GeoVectorDataset or MetDataset")
        self.set_source(source)

        if "rhi" in self.source:
            warnings.warn(
                "Variable 'rhi' already found on source to be scaled. This "
                "is unexpected and may be the result of humidity scaling "
                "being applied more than once."
            )

        if isinstance(self.source, GeoVectorDataset):
            self.source.setdefault("air_pressure", self.source.air_pressure)

        q = self.source.data["specific_humidity"]
        T = self.source.data["air_temperature"]
        p = self.source.data["air_pressure"]
        kwargs = {k: self.get_source_param(k) for k in self.scaler_specific_keys}

        q, rhi = self.scale(q, T, p, **kwargs)
        self.source.update(specific_humidity=q, rhi=rhi)

        return self.source


@dataclasses.dataclass
class ConstantHumidityScalingParams(ModelParams):
    """Parameters for :class:`ConstantHumidityScaling`."""

    #: Scale specific humidity by dividing it with adjustment factor per
    #: :cite:`schumannContrailCirrusPrediction2012` eq. (9). Set to a constant
    #: 0.9 in :cite:`schumannContrailCirrusPrediction2012` to account for sub-scale
    #: variability of specific humidity. A value of 1.0 provides no scaling.
    rhi_adj: float = 0.97


class ConstantHumidityScaling(HumidityScaling):
    """Scale specific humidity by applying a constant uniform scaling.

    This scalar simply applies the transformation..

        rhi -> rhi / rhi_adj

    where ``rhi_adj`` is a constant specified by :attr:`params` or overridden by
    a variable or attribute on ``source`` in :meth:`eval`.

    The convention to divide by ``rhi_adj`` instead of considering the more natural
    product ``rhi_adj * rhi`` is somewhat arbitrary. In short, ``rhi_adj`` can be
    thought of as the critical threshold for supersaturation.

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    - :cite:`reutterIceSupersaturatedRegions2020`
    """

    name = "constant_scale"
    long_name = "Constant humidity scaling"
    formula = "rhi -> rhi / rhi_adj"
    default_params = ConstantHumidityScalingParams
    scaler_specific_keys = ("rhi_adj",)

    @overrides
    def scale(
        self,
        specific_humidity: ArrayLike,
        air_temperature: ArrayLike,
        air_pressure: ArrayLike,
        **kwargs: Any,
    ) -> tuple[ArrayLike, ArrayLike]:
        rhi_adj = kwargs.get("rhi_adj", self.params["rhi_adj"])
        q = specific_humidity / rhi_adj
        rhi = thermo.rhi(q, air_temperature, air_pressure)
        return q, rhi


@dataclasses.dataclass
class ExponentialBoostHumidityScalingParams(ConstantHumidityScalingParams):
    """Parameters for :class:`ExponentialBoostHumidityScaling`."""

    #: Boost RHi values exceeding 1 as described in :cite:`teohAviationContrailClimate2022`.
    #: In :meth:`eval`, this can be overridden by a keyword argument with the same name.
    rhi_boost_exponent: float = 1.7

    #: Used to clip overinflated unrealistic RHi values.
    clip_upper: float = 1.7


class ExponentialBoostHumidityScaling(HumidityScaling):
    """Scale humidity by composing constant scaling with exponential boosting.

    This formula composes the transformations

    #. constant scaling: ``rhi -> rhi / rhi_adj``
    #. exponential boosting: ``rhi -> rhi ^ rhi_boost_exponent if rhi > 1``
    #. clipping: ``rhi -> min(rhi, clip_upper)``

    where ``rhi_adj``, ``rhi_boost_exponent``, and ``clip_upper`` are model :attr:`params`.

    References
    ----------
    - :cite:`teohAviationContrailClimate2022`

    See Also
    --------
    :class:`ExponentialBoostLatitudeCorrectionHumidityScaling`
    """

    name = "exponential_boost"
    long_name = "Constant humidity scaling composed with exponential boosting"
    formula = "rhi -> (rhi / rhi_adj) ^ rhi_boost_exponent"
    default_params = ExponentialBoostHumidityScalingParams
    scaler_specific_keys = "rhi_adj", "rhi_boost_exponent", "clip_upper"

    @overrides
    def scale(
        self,
        specific_humidity: ArrayLike,
        air_temperature: ArrayLike,
        air_pressure: ArrayLike,
        **kwargs: Any,
    ) -> tuple[ArrayLike, ArrayLike]:
        # Get coefficients
        rhi_adj = kwargs["rhi_adj"]
        rhi_boost_exponent = kwargs["rhi_boost_exponent"]
        clip_upper = kwargs["clip_upper"]

        # Compute uncorrected RHi
        rhi_over_q = _rhi_over_q(air_temperature, air_pressure)
        rhi = specific_humidity * rhi_over_q

        # Correct RHi
        rhi /= rhi_adj

        # Find ISSRs
        is_issr = rhi >= 1

        # Apply boosting to ISSRs
        if isinstance(rhi, xr.DataArray):
            rhi = rhi.where(~is_issr, rhi**rhi_boost_exponent)
            rhi = rhi.clip(max=clip_upper)

        else:
            boosted_rhi = rhi[is_issr] ** rhi_boost_exponent
            boosted_rhi.clip(max=clip_upper, out=boosted_rhi)
            rhi[is_issr] = boosted_rhi

        # Recompute specific_humidity from corrected rhi
        specific_humidity = rhi / rhi_over_q

        # Return the pair
        return specific_humidity, rhi


@dataclasses.dataclass
class ExponentialBoostLatitudeCorrectionHumidityScalingParams(ModelParams):
    """Parameters for :class:`ExponentialBoostLatitudeCorrectionHumidityScaling`."""

    #: Constants obtained by fitting a sigmoid curve to IAGOS data. These can be
    #: overridden by keyword arguments with the same name.
    rhi_a0: float = 0.062621
    rhi_a1: float = 0.45893
    rhi_a2: float = 39.254
    rhi_a3: float = 0.95224
    rhi_b0: float = 1.4706
    rhi_b1: float = 0.44312
    rhi_b2: float = 18.755
    rhi_b3: float = 1.4325


class ExponentialBoostLatitudeCorrectionHumidityScaling(HumidityScaling):
    """Correct RHi values derived from ECMWF ERA5 HRES.

    Unlike other RHi correction factors, this function applies a custom latitude-based
    term and has been tuned for global application.

    This formula composes the transformations

    #. constant scaling: ``rhi -> rhi / rhi_adj``
    #. exponential boosting: ``rhi -> rhi ^ rhi_boost_exponent if rhi > 1``
    #. clipping: ``rhi -> min(rhi, rhi_max)``

    where ``rhi_adj`` and ``rhi_boost_exponent`` depend on ``latitude`` to minimize
    error between ERA5 HRES and IAGOS in-situ data.

    For each waypoint, ``rhi_max`` ensures that the corrected RHi does not exceed the
    maximum value according to thermodynamics:

    - ``rhi_max = p_liq(T) / p_ice(T)`` for ``T > 235 K``,
      (Pruppacher and Klett, 1997)
    - ``rhi_max = 1.67 + (1.45 - 1.67) * (T - 190.) / (235. - 190.)`` for ``T < 235 K``
      (Karcher and Lohmann, 2002; Tompkins et al., 2007)

    The RHi correction addresses the known limitations of the ERA5 HRES humidity fields,
    ensuring that the ISSR coverage area and RHi-distribution is consistent with in-situ
    measurements from the IAGOS dataset. Generally, the correction:

    #. reduces the ISSR coverage area near the equator,
    #. increases the ISSR coverage area at higher latitudes, and
    #. accounts for localized regions with very high ice supersaturation (RHi > 120%).

    This methodology is an extension of Teoh et al. (2022) and has not yet been
    peer-reviewed/published.

    The ERA5 HRES <> IAGOS fitting uses a sigmoid curve to capture significant
    changes in tropopause height at 20 - 50 degrees latitude.

    The method :meth:`eval` requires a ``latitude`` keyword argument.

    References
    ----------
    - :cite:`teohAviationContrailClimate2022`
    - Kärcher, B. and Lohmann, U., 2002. A parameterization of cirrus cloud formation: Homogeneous
      freezing of supercooled aerosols. Journal of Geophysical Research: Atmospheres, 107(D2),
      pp.AAC-4.
    - Pruppacher, H.R. and Klett, J.D. (1997) Microphysics of Clouds and Precipitation. 2nd Edition,
      Kluwer Academic, Dordrecht, 954 p.
    - Tompkins, A.M., Gierens, K. and Rädel, G., 2007. Ice supersaturation in the ECMWF integrated
      forecast system. Quarterly Journal of the Royal Meteorological Society: A journal of the
      atmospheric sciences, applied meteorology and physical oceanography, 133(622), pp.53-63.

    See Also
    --------
    :class:`ExponentialBoostHumidityScaling`
    """

    name = "exponential_boost_latitude_customization"
    long_name = "Latitude specific humidity scaling composed with exponential boosting"
    formula = "rhi -> (rhi / rhi_adj) ^ rhi_boost_exponent"
    default_params = ExponentialBoostLatitudeCorrectionHumidityScalingParams
    scaler_specific_keys = (
        "latitude",
        "rhi_a0",
        "rhi_a1",
        "rhi_a2",
        "rhi_a3",
        "rhi_b0",
        "rhi_b1",
        "rhi_b2",
        "rhi_b3",
    )

    @overrides
    def scale(
        self,
        specific_humidity: ArrayLike,
        air_temperature: ArrayLike,
        air_pressure: ArrayLike,
        **kwargs: Any,
    ) -> tuple[ArrayLike, ArrayLike]:
        # Get sigmoid coefficients
        a0 = kwargs["rhi_a0"]
        a1 = kwargs["rhi_a1"]
        a2 = kwargs["rhi_a2"]
        a3 = kwargs["rhi_a3"]
        b0 = kwargs["rhi_b0"]
        b1 = kwargs["rhi_b1"]
        b2 = kwargs["rhi_b2"]
        b3 = kwargs["rhi_b3"]

        # Use the dtype of specific_humidity to determine the precision of the
        # the calculation. If working with gridded data here, latitude will have
        # float64 precision.
        latitude = kwargs["latitude"]
        if latitude.dtype != specific_humidity.dtype:
            latitude = latitude.astype(specific_humidity.dtype)
        lat_abs = np.abs(latitude)

        # Compute uncorrected RHi
        rhi_over_q = _rhi_over_q(air_temperature, air_pressure)
        rhi = specific_humidity * rhi_over_q

        # Calculate the rhi_adj factor and correct RHi
        rhi_adj = a0 / (1.0 + np.exp(a1 * (lat_abs - a2))) + a3
        rhi /= rhi_adj

        # Find ISSRs
        is_issr = rhi >= 1

        # Limit RHi to maximum value allowed by physics
        rhi_max = _calc_rhi_max(air_temperature)

        # Apply boosting to ISSRs
        if isinstance(rhi, xr.DataArray):
            boost_exponent = b0 / (1.0 + np.exp(b1 * (lat_abs - b2))) + b3
            rhi = rhi.where(~is_issr, rhi**boost_exponent)
            rhi = rhi.clip(max=rhi_max)

        else:
            # Calculate the optimal b coefficient over points in ISSRs
            boost_exponent = b0 / (1.0 + np.exp(b1 * (lat_abs[is_issr] - b2))) + b3

            # Apply boosting to ISSRs
            rhi[is_issr] = rhi[is_issr] ** boost_exponent
            rhi.clip(max=rhi_max, out=rhi)

        # Recompute specific_humidity from corrected rhi
        specific_humidity = rhi / rhi_over_q

        # Return the pair
        return specific_humidity, rhi


def _calc_rhi_max(air_temperature: ArrayLike) -> ArrayLike:
    p_ice: ArrayLike
    p_liq: ArrayLike

    if isinstance(air_temperature, xr.DataArray):
        p_ice = thermo.e_sat_ice(air_temperature)
        p_liq = thermo.e_sat_liquid(air_temperature)
        return xr.where(
            air_temperature < 235,
            1.67 + (1.45 - 1.67) * (air_temperature - 190.0) / (235.0 - 190.0),
            p_liq / p_ice,
        )

    low = air_temperature < 235
    air_temperature_low = air_temperature[low]
    air_temperature_high = air_temperature[~low]

    p_ice = thermo.e_sat_ice(air_temperature_high)
    p_liq = thermo.e_sat_liquid(air_temperature_high)

    out = np.empty_like(air_temperature)
    out[low] = 1.67 + (1.45 - 1.67) * (air_temperature_low - 190.0) / (235.0 - 190.0)
    out[~low] = p_liq / p_ice
    return out


@dataclasses.dataclass
class HumidityScalingByLevelParams(ModelParams):
    """Parameters for :class:`HumidityScalingByLevel`."""

    #: Fraction of troposphere for mid-troposphere humidity scaling.
    #: Default value suggested in :cite:`schumannContrailCirrusPrediction2012`.
    rhi_adj_mid_troposphere: float = 0.8

    #: Fraction of troposphere for stratosphere humidity scaling.
    #: Default value suggested in :cite:`schumannContrailCirrusPrediction2012`.
    rhi_adj_stratosphere: float = 1.0

    #: Adjustment factor for mid-troposphere humidity scaling. Default value
    #: of 0.8 taken from :cite:`schumannContrailCirrusPrediction2012`.
    mid_troposphere_threshold: float = 0.8

    #: Adjustment factor for stratosphere humidity scaling. Default value
    #: of 1.0 taken from :cite:`schumannContrailCirrusPrediction2012`.
    stratosphere_threshold: float = 1.0


class HumidityScalingByLevel(HumidityScaling):
    """Apply custom scaling to specific_humidity by pressure level.

    This implements the original humidity scaling scheme suggested in
    :cite:`schumannContrailCirrusPrediction2012`. In particular, see eq. (9)
    and the surrounding text, quoted below.

        Hence, the critical value RHic is usually taken different and below 100%
        in NWP models. In the ECMWF model, this value is..

            RHic = 0.8, (9)

        in the mid-troposphere, 1.0 in the stratosphere and follows
        a smooth transition with pressure altitude between these two
        values in the upper 20 % of the troposphere. For simplicity
        of further analysis, we divide the input value of q by RHic
        initially.

    See :class:`ConstantHumidityScaling` for the simple method described
    above.

    The diagram below shows the piecewise-linear ``rhi_adj`` factor by
    level. In particular, ``rhi_adj`` is constant at the stratosphere and above,
    linearly changes from the mid-troposphere to the stratosphere, and is
    constant at the mid-troposphere and below.

    ::

                     _________  stratosphere rhi_adj = 1.0
                    /
                   /
                  /
        _________/  mid-troposphere rhi_adj = 0.8

    References
    ----------
    - :cite:`schumannContrailCirrusPrediction2012`
    """

    name = "constant_scale_by_level"
    long_name = "Constant humidity scaling by level"
    formula = "rhi -> rhi / rhi_adj"
    default_params = HumidityScalingByLevelParams
    scaler_specific_keys = (
        "rhi_adj_mid_troposphere",
        "rhi_adj_stratosphere",
        "mid_troposphere_threshold",
        "stratosphere_threshold",
    )

    @overrides
    def scale(
        self,
        specific_humidity: ArrayLike,
        air_temperature: ArrayLike,
        air_pressure: ArrayLike,
        **kwargs: Any,
    ) -> tuple[ArrayLike, ArrayLike]:
        rhi_adj_mid_troposphere = kwargs["rhi_adj_mid_troposphere"]
        rhi_adj_stratosphere = kwargs["rhi_adj_stratosphere"]
        mid_troposphere_threshold = kwargs["mid_troposphere_threshold"]
        stratosphere_threshold = kwargs["stratosphere_threshold"]

        thresholds = np.array([stratosphere_threshold, mid_troposphere_threshold])
        xp = units.m_to_pl(constants.h_tropopause * thresholds)  # type: ignore

        # np.interp expects a sorted parameter `xp`
        # The calculation will get bungled if this isn't the case
        if xp[0] > xp[1]:
            raise ValueError(
                "Attribute 'stratosphere_threshold' must exceed "
                "attribute 'mid_troposphere_threshold'."
            )

        level = air_pressure / 100
        fp = [rhi_adj_stratosphere, rhi_adj_mid_troposphere]
        rhi_adj = np.interp(level, xp=xp, fp=fp)

        q = specific_humidity / rhi_adj
        rhi = thermo.rhi(q, air_temperature, air_pressure)
        return q, rhi



"""Test `pycontrails.core.datalib` module."""

from __future__ import annotations

from datetime import datetime

import numpy as np
import pandas as pd
import pytest

from pycontrails.core import datalib
from pycontrails.core.met_var import AirTemperature, EastwardWind, NorthwardWind, RelativeHumidity


def test_parse_timesteps() -> None:
    """Test timestep parsing."""

    # accept single time
    ts = datalib.parse_timesteps(datetime(2019, 5, 31, 0))
    assert ts == [datetime(2019, 5, 31, 0)]

    ts = datalib.parse_timesteps([datetime(2019, 5, 31, 0)])
    assert ts == [datetime(2019, 5, 31, 0)]

    # select floor and ceiling hour for single time with minutes defined
    ts = datalib.parse_timesteps(datetime(2019, 5, 31, 5, 10))
    assert ts == [datetime(2019, 5, 31, 5), datetime(2019, 5, 31, 6)]

    # throw ValueError for length == 0
    with pytest.raises(ValueError):
        datalib.parse_timesteps([])

    # throw ValueError for length > 2
    with pytest.raises(ValueError):
        datalib.parse_timesteps(
            [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 0)]
        )

    # accept (start, end)
    ts = datalib.parse_timesteps((datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 3)))
    assert ts == [
        datetime(2019, 5, 31, 0),
        datetime(2019, 5, 31, 1),
        datetime(2019, 5, 31, 2),
        datetime(2019, 5, 31, 3),
    ]

    ts = datalib.parse_timesteps((datetime(2019, 5, 31, 0, 29), datetime(2019, 5, 31, 2, 40)))
    assert ts == [
        datetime(2019, 5, 31, 0),
        datetime(2019, 5, 31, 1),
        datetime(2019, 5, 31, 2),
        datetime(2019, 5, 31, 3),
    ]

    # support alternate types for input
    ts = datalib.parse_timesteps(pd.to_datetime(datetime(2019, 5, 31, 0, 29)))
    assert ts == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1)]

    ts = datalib.parse_timesteps("2019-05-31T00:29:00")
    assert ts == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1)]
    ts = datalib.parse_timesteps(("2019-05-31T00:29:00", "2019-05-31T01:40:00"))
    assert ts == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1), datetime(2019, 5, 31, 2)]

    ts = datalib.parse_timesteps(np.datetime64("2019-05-31T00:29:00"))
    assert ts == [datetime(2019, 5, 31, 0), datetime(2019, 5, 31, 1)]
    ts = datalib.parse_timesteps(
        (np.datetime64("2019-05-31T00:29:00"), np.datetime64("2019-05-31T02:40:00"))
    )
    assert ts == [
        datetime(2019, 5, 31, 0),
        datetime(2019, 5, 31, 1),
        datetime(2019, 5, 31, 2),
        datetime(2019, 5, 31, 3),
    ]

    # select frequency
    ts = datalib.parse_timesteps(("2019-05-31T00:29:00", "2019-05-31T08:40:00"), freq="3H")
    assert ts == [
        datetime(2019, 5, 31, 0),
        datetime(2019, 5, 31, 3),
        datetime(2019, 5, 31, 6),
        datetime(2019, 5, 31, 9),
    ]

    # set freq to None to get back input as a list
    ts = datalib.parse_timesteps("2019-05-31T00:29:00", freq=None)
    assert ts == [datetime(2019, 5, 31, 0, 29), datetime(2019, 5, 31, 0, 29)]
    ts = datalib.parse_timesteps(("2019-05-31T00:29:00", "2019-05-31T01:40:00"), freq=None)
    assert ts == [datetime(2019, 5, 31, 0, 29), datetime(2019, 5, 31, 1, 40)]


def test_parse_pressure_levels() -> None:
    """Test pressure level parsing."""

    # if input is -1, return [-1]
    pl = datalib.parse_pressure_levels([-1])
    assert pl == [-1]

    # support single values
    pl = datalib.parse_pressure_levels(100)
    assert pl == [100]

    # cast floats to int
    pl = datalib.parse_pressure_levels(100.0)
    assert pl == [100]

    pl = datalib.parse_pressure_levels([100.0, 200])
    assert pl == [100, 200]

    # support ndarrays
    pl = datalib.parse_pressure_levels(np.array([100, 200]))
    assert pl == [100, 200]

    # support ndarrays with -1
    pl = datalib.parse_pressure_levels(np.array([-1, 200]))
    assert pl == [-1, 200]

    # throw error if not supported
    with pytest.raises(ValueError, match="not a valid pressure level"):
        datalib.parse_pressure_levels([100, 200, 300], supported=[100, 200])


def test_parse_variables() -> None:
    """Test variable parsing."""

    supported = [AirTemperature, RelativeHumidity]

    # by MetVariable
    v = datalib.parse_variables(AirTemperature, supported=supported)
    assert v == [AirTemperature]

    # raise value error if unmatched
    with pytest.raises(ValueError, match="not in supported parameters"):
        datalib.parse_variables(EastwardWind, supported=supported)

    # by list[MetVariable]
    v = datalib.parse_variables(
        [[EastwardWind, AirTemperature, RelativeHumidity], RelativeHumidity], supported=supported
    )
    assert v == [AirTemperature, RelativeHumidity]

    # raise value error if unmatched
    with pytest.raises(ValueError, match="not in supported parameters"):
        datalib.parse_variables([[EastwardWind, NorthwardWind]], supported=supported)

    # raise value error if a str
    with pytest.raises(ValueError, match="must be of type MetVariable"):
        datalib.parse_variables([["eastward_wind", "northward_wind"]], supported=supported)

    # by short name
    v = datalib.parse_variables("t", supported=supported)
    assert v == [AirTemperature]
    v = datalib.parse_variables(["t"], supported=supported)
    assert v == [AirTemperature]
    v = datalib.parse_variables(["t", "r"], supported=supported)
    assert v == [AirTemperature, RelativeHumidity]

    # by standard name
    v = datalib.parse_variables("air_temperature", supported=supported)
    assert v == [AirTemperature]
    v = datalib.parse_variables(["air_temperature"], supported=supported)
    assert v == [AirTemperature]

    # by grib id
    v = datalib.parse_variables(11, supported=supported)
    assert v == [AirTemperature]
    v = datalib.parse_variables([11], supported=supported)
    assert v == [AirTemperature]

    # by ecmwf id
    v = datalib.parse_variables(130, supported=supported)
    assert v == [AirTemperature]
    v = datalib.parse_variables([130], supported=supported)
    assert v == [AirTemperature]

    # support ndarray input
    v = datalib.parse_variables(np.array(["t", "r"]), supported=supported)
    assert v == [AirTemperature, RelativeHumidity]

    # make sure that original dataclass is copied
    v = datalib.parse_variables("t", supported=supported)
    assert v[0] is not AirTemperature

    # raise value error if unmatched
    with pytest.raises(ValueError, match="not in supported parameters"):
        datalib.parse_variables("vorticity", supported=supported)


def test_parse_grid() -> None:
    """Test grid parsing."""

    # raise value error if not supported
    with pytest.raises(ValueError, match="input must be one of"):
        datalib.parse_grid(0.3, supported=[0.25, 0.5])


def test_met_datasource() -> None:
    """Test abstract MetDataSource class."""
    pass


def test_round_hour() -> None:
    """Test round hour helper."""

    time = datetime(2022, 12, 7, 14)
    rounded = datalib.round_hour(time, 2)
    assert rounded == time

    time = datetime(2022, 12, 7, 15)
    rounded = datalib.round_hour(time, 2)
    assert rounded == datetime(2022, 12, 7, 14)

    time = datetime(2022, 12, 7, 17)
    rounded = datalib.round_hour(time, 6)
    assert rounded == datetime(2022, 12, 7, 12)

    time = datetime(2022, 12, 7, 17)
    rounded = datalib.round_hour(time, 1)
    assert rounded == datetime(2022, 12, 7, 17)

    with pytest.raises(ValueError, match="must be between"):
        _ = datalib.round_hour(time, 25)



"""
``pycontrails`` public API.

Copyright 2021-2023 Breakthrough Energy

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

import logging
from importlib import metadata

import dask

# Work around for https://github.com/pydata/xarray/issues/7259
# Only occurs for xarray 2022.11 and above
try:
    import netCDF4
except ModuleNotFoundError:
    pass

from pycontrails.core.cache import DiskCacheStore, GCPCacheStore
from pycontrails.core.datalib import MetDataSource
from pycontrails.core.fleet import Fleet
from pycontrails.core.flight import Aircraft, Flight
from pycontrails.core.fuel import Fuel, HydrogenFuel, JetA, SAFBlend
from pycontrails.core.met import MetDataArray, MetDataset
from pycontrails.core.met_var import MetVariable
from pycontrails.core.models import Model, ModelParams
from pycontrails.core.vector import GeoVectorDataset, VectorDataset

__version__ = metadata.version("pycontrails")
__license__ = "Apache-2.0"
__url__ = "https://py.contrails.org"

log = logging.getLogger(__name__)

# Hardcode the dask warning silence config
dask.config.set({"array.slicing.split_large_chunks": False})


__all__ = [
    "Aircraft",
    "DiskCacheStore",
    "Fleet",
    "Flight",
    "Fuel",
    "GCPCacheStore",
    "GeoVectorDataset",
    "HydrogenFuel",
    "JetA",
    "MetDataArray",
    "MetDataset",
    "MetDataSource",
    "MetVariable",
    "Model",
    "ModelParams",
    "SAFBlend",
    "VectorDataset",
]



"""Test vector module."""

from __future__ import annotations

import warnings

import numpy as np
import pandas as pd
import pytest

from pycontrails import GeoVectorDataset, MetDataset, VectorDataset
from pycontrails.core.vector import AttrDict, VectorDataDict


@pytest.fixture(scope="module")
def random_path() -> VectorDataset:
    """Return meaningless `VectorDataset` fixture."""
    rng = np.random.default_rng(123)
    data = {
        "a": rng.random(100),
        "b": rng.random(100),
        "c": np.arange(100),
    }
    attrs = {"test": "attribute"}
    return VectorDataset(data=data, attrs=attrs, copy=False)


@pytest.fixture(scope="module")
def random_geo_path() -> GeoVectorDataset:
    """Demo GeoVectorDataset.

    Returns
    -------
    GeoVectorDataset
    """
    data = {
        "longitude": np.array([-15, -14, -13, -12]),
        "latitude": np.array([10, 11, 12, 13]),
        "time": np.array(
            [
                "2021-10-05 00:00:00",
                "2021-10-05 01:00:00",
                "2021-10-05 02:00:00",
                "2021-10-05 03:00:00",
            ]
        ),
        "altitude": np.array([1000, 5000, 10000, 40000]),
        "a": np.array([1, 2, 3, 4]),
        "b": np.array([5, 6, 7, 8]),
    }
    attrs = {"test": "attribute"}
    with warnings.catch_warnings():
        # Pandas has no trouble doing the conversion
        warnings.filterwarnings("ignore", "Time data is not np.datetime64")
        return GeoVectorDataset(data=data, attrs=attrs, copy=False)


def test_vector_init() -> None:
    """VectorDataset __init__."""

    # test constructor
    data = {
        "a": np.ones(100),
        "b": np.empty(100),
    }
    attrs = {
        "test": "value",
    }
    vds = VectorDataset(data, attrs=attrs)
    assert isinstance(vds.data, VectorDataDict)
    assert isinstance(vds.attrs, AttrDict)

    # copy by default
    assert vds.data["a"] is not data["a"]
    assert np.all(vds.data["a"] == data["a"])

    assert vds.attrs is not attrs
    assert vds.attrs["test"] == attrs["test"]

    # it will acceptVectorDataDict/AttrDict input, but still copy
    data = VectorDataDict(
        {
            "a": np.ones(100),
            "b": np.empty(100),
        }
    )
    attrs = AttrDict(
        {
            "test": "value",
        }
    )
    vds = VectorDataset(data, attrs=attrs)
    assert vds.data is not data
    assert vds.attrs is not attrs

    vds = VectorDataset(data, attrs=attrs, copy=False)
    assert vds.data is data
    assert vds.attrs is attrs


def test_bad_size_pass_data() -> None:
    """Ensure an error is raised when inconsistent sizes are passed in the constructor."""
    data = {
        "a": np.arange(5),
        "b": np.arange(6),
        "c": np.arange(5),
    }
    with pytest.raises(ValueError, match="sizes"):
        VectorDataset(data)

    del data["b"]
    vds1 = VectorDataset(data)
    assert vds1.size == 5

    vds2 = VectorDataset()
    vds2["a"] = data["a"]
    vds2["c"] = data["c"]
    assert vds2.size == 5
    assert vds1 == vds2


def test_vector_empty() -> None:
    """Confirm an empty `VectorDataset` instance behaves as expected."""
    attrs = {"test": "attribute"}
    vds = VectorDataset(attrs=attrs)

    assert vds.data == {}
    assert vds.attrs["test"] == "attribute"
    assert vds.size == 0
    assert "0 keys" in repr(vds)
    assert "VectorDataset" in repr(vds)

    vds["orcas"] = np.ones(10)
    assert vds
    assert vds.size == 10
    assert "orcas" in repr(vds)
    assert "VectorDataset" in repr(vds)

    # deleting that last key sets the size back to 0
    del vds["orcas"]
    assert vds.size == 0


def test_vector_dict_like(random_path: VectorDataset) -> None:
    """Test vector dictlike methods"""
    vds = random_path.copy()

    # test default dict methods
    # get
    assert vds["a"] is vds.data["a"]
    assert vds.get("a", None) is vds.data["a"]
    assert vds.get("d", None) is None

    # set
    d = np.random.random(vds["a"].size)
    vds["d"] = d
    assert vds.data["d"] is d
    with pytest.raises(ValueError, match="sizes"):
        vds["e"] = np.empty(5)

    # del
    del vds["d"]
    assert "d" not in vds.data

    # update
    vds.update({"d": d})
    assert vds.data["d"] is d
    vds.update(e=d)
    assert vds.data["e"] is d
    # (`kwargs`` overwrites `other``)
    vds.update({"d": np.zeros_like(d)}, e=np.ones_like(d), d=np.ones_like(d))
    assert np.all(vds.data["d"] == 1)
    assert np.all(vds.data["e"] == 1)

    with pytest.raises(ValueError, match="sizes"):
        vds.update(d=np.empty(5))
    with pytest.raises(ValueError, match="sizes"):
        vds.data.update(d=np.empty(5))

    # setdefault
    out = vds.setdefault("d", d)
    assert np.all(vds.data["d"] == 1)
    assert vds.data["d"] is out
    out = vds.setdefault("f", d)
    assert np.all(vds.data["f"] == d)
    assert vds.data["f"] is out

    with pytest.raises(ValueError, match="sizes"):
        out = vds.setdefault("g", np.empty(5))

    with pytest.raises(ValueError, match="sizes"):
        out = vds.setdefault("g")

    # setdefault overwrites None values
    # out = vds["g"] = None

    # iter
    assert list(vds) == list(vds.data)


def test_vector_attrs(random_path: VectorDataset) -> None:
    """Vector attr tests."""
    vds = random_path.copy()

    # setdefault has special behavior to overwrite None
    vds.attrs["none"] = None
    default = vds.attrs.setdefault("none", "value")
    assert vds.attrs["none"] == "value" and default == "value"


def test_vector_not_dict_like(random_path: VectorDataset) -> None:
    """Test un-dict-like methods."""

    vds = random_path.copy()

    # len, size, shape
    assert len(vds) == 100
    assert vds.size == 100
    assert vds.shape == (100,)

    # bool
    # when all data is deleted, the size should be reset
    del vds["a"]
    del vds["b"]
    assert len(vds) == 100
    assert vds

    del vds["c"]
    assert not vds
    assert len(vds) == 0
    assert vds.size == 0

    # eq
    vds = random_path.copy()
    vds2 = random_path.copy()
    assert vds == vds2
    vds.update(a=np.ones_like(vds["a"]))
    assert vds != vds2


def test_vector_repr() -> None:
    """Confirm `VectorDataset.__repr__` clips very long attributes."""
    attrs = {"short_key": "short_val", "long_key": "long_val" * 100}
    vds = VectorDataset(attrs=attrs)
    out = repr(vds)

    assert "short_key" in out
    assert "long_key" in out
    assert "short_val" in out

    # The very long value gets truncated
    assert out.endswith("...")
    assert "long_val" * 5 in out
    assert "long_val" * 10 not in out


def test_warning_on_set(random_path: VectorDataset) -> None:
    """Test the warning on __setitem__() and set_attr()"""

    # should not warn on empty and the same value object
    vds = VectorDataset()
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        val = np.ones(10)
        vds["orcas"] = val
        vds["orcas"] = val  # won't warn for the same object

        # shouldn't warn on data either
        vds.data["orcas"] = val

        # shouldn't warn when `update` is used
        vds.update(orcas=np.zeros(10))
        vds.update({"orcas": np.empty(10)})
        vds.data.update(orcas=np.zeros(10))

    # WILL warn for the same values, different object
    with pytest.warns(UserWarning, match="Overwriting data"):
        vds["orcas"] = np.ones(10)
    with pytest.warns(UserWarning, match="Overwriting data"):
        vds.data["orcas"] = np.ones(10)

    # WILL warn for different values too
    with pytest.warns(UserWarning, match="Overwriting data"):
        vds["orcas"] = np.zeros(10)
    with pytest.warns(UserWarning, match="Overwriting data"):
        vds.data["orcas"] = np.zeros(10)

    # should not warn on None attr and the same value object
    vds.attrs["test"] = None
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        val = "value"
        vds.attrs["test"] = val
        vds.attrs["test"] = val  # won't warn for the same object
        vds.attrs["test"] = "value"  # won't warn for the same object

        # should not warn with `update`
        vds.attrs.update(test="different")

    # should warn on overwrite attr
    with pytest.warns(UserWarning, match="Overwriting attr"):
        vds.attrs["test"] = "new_value"


def test_to_dataframe(random_path: VectorDataset, random_geo_path: GeoVectorDataset) -> None:
    """Test method `to_dataframe`."""
    df = random_path.to_dataframe()
    assert df.shape == (100, 3)
    assert df.columns.tolist() == list(random_path.data)

    # reverse gives same data dictionary
    data = df.to_dict(orient="list")
    assert random_path.data.keys() == data.keys()
    for arr1, arr2 in zip(random_path.data.values(), data.values()):
        np.testing.assert_array_equal(arr1, arr2)

    df = random_geo_path.to_dataframe()
    assert "latitude" in df
    assert "longitude" in df
    assert "time" in df
    assert "altitude" in df


def test_sort() -> None:
    """Test `.sort` method."""
    data = {
        "a": np.array([1, 10, 2, 2]),
        "b": np.array([1, 4, 3, 2]),
        "c": np.array([1, 2, 3, 10]),
    }
    attrs = {"test": "attribute"}
    vds = VectorDataset(data=data, attrs=attrs)
    sorted_vds = vds.sort("a")
    assert sorted_vds is not vds
    assert np.all(sorted_vds["a"] == np.array([1, 2, 2, 10]))
    assert np.all(sorted_vds["b"] == np.array([1, 3, 2, 4]))
    assert np.all(sorted_vds["c"] == np.array([1, 3, 10, 2]))

    sorted_vds = vds.sort(["a", "b"])
    assert sorted_vds is not vds
    assert np.all(sorted_vds["a"] == np.array([1, 2, 2, 10]))
    assert np.all(sorted_vds["b"] == np.array([1, 2, 3, 4]))
    assert np.all(sorted_vds["c"] == np.array([1, 10, 3, 2]))


def test_geovector_init() -> None:
    """Test GeoVectorDataset.__init__()."""
    longitude = np.linspace(0, 50, 100)
    latitude = np.linspace(0, 10, 100)
    time = pd.date_range("2019-01-01 00:00:00", "2019-01-01 02:00:00", periods=100)

    # with altitude
    altitude = np.linspace(11000, 11500, 100)
    gvds = GeoVectorDataset(longitude=longitude, latitude=latitude, altitude=altitude, time=time)
    assert gvds["longitude"] is not longitude
    assert gvds["latitude"] is not latitude
    assert gvds["altitude"] is not altitude
    assert gvds["time"] is not time
    assert np.all(gvds["longitude"] == longitude)
    assert np.all(gvds["latitude"] == latitude)
    assert np.all(gvds["altitude"] == altitude)
    assert np.all(gvds["time"] == time)
    assert "level" not in gvds

    # with level
    level = np.linspace(250, 300, 100)
    gvds = GeoVectorDataset(longitude=longitude, latitude=latitude, level=level, time=time)
    assert gvds["longitude"] is not longitude
    assert gvds["latitude"] is not latitude
    assert gvds["level"] is not level
    assert gvds["time"] is not time
    assert np.all(gvds["longitude"] == longitude)
    assert np.all(gvds["latitude"] == latitude)
    assert np.all(gvds["level"] == level)
    assert np.all(gvds["time"] == time)
    assert "altitude" not in gvds

    # dataframe
    df = pd.DataFrame()
    df["longitude"] = longitude
    df["latitude"] = latitude
    df["altitude"] = 11000
    df["time"] = time
    gvds = GeoVectorDataset(data=df)
    assert gvds["longitude"] is not longitude
    assert gvds["latitude"] is not latitude
    assert gvds["altitude"] is not altitude
    assert gvds["time"] is not time
    assert np.all(gvds["longitude"] == longitude)
    assert np.all(gvds["latitude"] == latitude)
    assert np.all(gvds["altitude"] == 11000)
    assert np.all(gvds["time"] == time)

    # mixed
    df = pd.DataFrame()
    df["longitude"] = longitude
    df["latitude"] = latitude
    gvds = GeoVectorDataset(data=df, altitude=altitude, time=time)
    assert np.all(gvds["longitude"] == longitude)
    assert np.all(gvds["latitude"] == latitude)
    assert np.all(gvds["altitude"] == altitude)
    assert np.all(gvds["time"] == time)

    # required keys
    df = pd.DataFrame()
    df["longitude"] = longitude
    df["latitude"] = latitude
    df["altitude"] = altitude
    df["time"] = time
    for col in ["time", "latitude", "longitude"]:
        data = df.drop(col, axis=1)
        with pytest.raises(KeyError, match="Missing required data keys:"):
            GeoVectorDataset(data=data)

    # fail with different lengths
    longitude = np.array([0, 1, 2])
    latitude = np.array([0, 1])
    altitude = np.array([11000, 11200])
    time = pd.date_range("2019-01-01 00:00:00", "2019-01-01 02:00:00", periods=2)
    df = pd.DataFrame()
    df["longitude"] = longitude
    with pytest.raises(ValueError, match="Incompatible array sizes: 2 and 3"):
        gvds = GeoVectorDataset(
            longitude=longitude, latitude=latitude, altitude=altitude, time=time
        )
    with pytest.raises(ValueError, match="Incompatible array sizes: 2 and 3"):
        gvds = GeoVectorDataset(data=df, latitude=latitude, altitude=altitude, time=time)

    # parse time
    longitude = np.array([0, 1])
    latitude = np.array([0, 1])
    altitude = np.array([11000, 11200])
    time = np.array(["2021-09-01 00:00:00", "2021-09-01 01:00:00"])

    with pytest.warns(UserWarning, match="not np.datetime64. Attempting to coerce"):
        gvds = GeoVectorDataset(
            longitude=longitude,
            latitude=latitude,
            altitude=altitude,
            time=time,
        )
    assert isinstance(gvds["time"][0], np.datetime64)
    assert np.all(gvds["longitude"] == longitude)
    assert np.all(gvds["latitude"] == latitude)
    assert np.all(gvds["altitude"] == altitude)

    # level / altitude keys
    longitude = np.linspace(0, 50, 100)
    latitude = np.linspace(0, 10, 100)
    time = pd.date_range("2019-01-01 00:00:00", "2019-01-01 02:00:00", periods=100)
    altitude = np.linspace(11000, 11500, 100)
    level = np.linspace(250, 300, 100)

    # See TODO in vector regarding level/altitude
    # with pytest.warns(Warning, match="contains both `level` and `altitude` keys"):
    #     gvds = GeoVectorDataset(
    #         longitude=longitude, latitude=latitude, level=level, altitude=altitude, time=time
    #     )

    match = "GeoVectorDataset requires either `level` or `altitude` or `altitude_ft` data key"
    with pytest.raises(KeyError, match=match):
        gvds = GeoVectorDataset(longitude=longitude, latitude=latitude, time=time)


def test_geovector_empty() -> None:
    """Test empty geovector handling."""

    gv = GeoVectorDataset()
    gv2 = GeoVectorDataset.create_empty()

    assert gv == gv2

    # TODO: what do you do with an empty GeoVectorDataset()
    # Do we need a special `update` so you can set the coordinates?
    # these works but could be error prone or clunky
    gv["longitude"] = np.linspace(0, 50, 100)
    assert gv.size == 100

    gv.update(
        latitude=np.linspace(0, 10, 100),
        longitude=np.linspace(0, 50, 100),
        altitude=np.linspace(11000, 11500, 100),
        time=pd.date_range("2019-01-01 00:00:00", "2019-01-01 02:00:00", periods=100),
    )
    assert gv.size == 100


def test_geovector_properties() -> None:
    """Test GeoVectorDataset @property()."""
    longitude = np.linspace(0, 50, 100)
    latitude = np.linspace(0, 10, 100)
    altitude = np.linspace(11000, 11500, 100)
    time = pd.date_range("2019-01-01 00:00:00", "2019-01-01 02:00:00", periods=100)

    gvds = GeoVectorDataset(longitude=longitude, latitude=latitude, altitude=altitude, time=time)
    assert gvds

    level = np.linspace(250, 300, 100)
    gvds = GeoVectorDataset(longitude=longitude, latitude=latitude, level=level, time=time)
    assert gvds["longitude"] is not longitude
    assert gvds["latitude"] is not latitude
    assert gvds["level"] is not level
    assert gvds["time"] is not time
    assert np.all(gvds["longitude"] == longitude)
    assert np.all(gvds["latitude"] == latitude)
    assert np.all(gvds["level"] == level)
    assert np.all(gvds["time"] == time)
    assert "altitude" not in gvds


def test_coord_intersect_met(met_issr: MetDataset) -> None:
    """Test coord_intersect_met."""

    longitude = np.linspace(0, 50, 100)
    latitude = np.linspace(0, 10, 100)
    level = np.linspace(230, 250, 100)
    time = pd.date_range("2019-05-31 05:01:00", "2019-05-31 05:50:00", periods=100)

    gvds = GeoVectorDataset(longitude=longitude, latitude=latitude, level=level, time=time)
    assert gvds.coords_intersect_met(met_issr).all()

    gvds = GeoVectorDataset(
        longitude=longitude, latitude=latitude, level=np.linspace(200, 250, 100), time=time
    )
    assert gvds.coords_intersect_met(met_issr).any()

    gvds = GeoVectorDataset(
        longitude=longitude,
        latitude=latitude,
        level=level,
        time=pd.date_range("2019-05-30 21:00:00", "2019-05-31 05:50:00", periods=100),
    )
    assert gvds.coords_intersect_met(met_issr).any()

    gvds = GeoVectorDataset(
        longitude=longitude, latitude=latitude, level=np.linspace(200, 220, 100), time=time
    )
    assert not gvds.coords_intersect_met(met_issr).any()


def test_broadcast_attrs(random_path: VectorDataset):
    """Test method `broadcast_attrs`."""
    path = random_path.copy()

    with pytest.raises(KeyError, match="does not contain attr `not-in-attr`"):
        path.broadcast_attrs("not-in-attr")

    assert len(path.attrs) == 1
    assert len(path.data.keys()) == 3
    path.broadcast_attrs("test")
    assert len(path.data.keys()) == 4
    assert np.all(path["test"] == "attribute")

    path.attrs.update(test="newattr")
    with pytest.warns(match="Found duplicate key"):
        path.broadcast_attrs("test")

    path.broadcast_attrs("test", overwrite=True)
    assert np.all(path["test"] == "newattr")


def test_broadcast_numeric(random_path: VectorDataset):
    """Test method `broadcast_numeric_attrs`."""
    path = random_path.copy()
    assert path.attrs is not random_path.attrs
    path.attrs["numeric"] = 5
    assert len(path.attrs) == 2
    assert len(random_path.attrs) == 1

    assert len(path.data.keys()) == 3
    path.broadcast_numeric_attrs()
    assert len(path.data.keys()) == 4
    assert "numeric" in path
    assert np.all(path["numeric"] == 5)

    # overwrite and ignore
    path.attrs.update(numeric=1)
    path.attrs["numeric_ignored"] = 6
    path.broadcast_numeric_attrs(ignore_keys="numeric_ignored", overwrite=True)
    assert len(path.data.keys()) == 4
    assert "numeric_ignored" not in path
    assert np.all(path["numeric"] == 1)

    # ignore str
    path = random_path.copy()
    path.attrs["numeric_ignored"] = 6
    path.broadcast_numeric_attrs(ignore_keys="numeric_ignored")
    assert len(path.data.keys()) == 3
    assert "numeric_ignored" not in path

    # ignore list
    path.broadcast_numeric_attrs(ignore_keys=["numeric_ignored"])
    assert len(path.data.keys()) == 3
    assert "numeric_ignored" not in path


def test_vector_copy(random_path: VectorDataset):
    """Confirm various "copy" patterns work as expected."""
    # Without copying, the resulting array is a view of the original array
    df = random_path.to_dataframe(copy=False)
    a = df["a"].to_numpy()
    assert a.base is random_path["a"]
    assert np.may_share_memory(a, random_path["a"])

    # Create a new VectorDataset and show that copying is still respected
    random_path2 = VectorDataset(df, copy=False)
    assert random_path2 == random_path
    assert random_path2["a"].base is random_path["a"]
    assert np.may_share_memory(random_path2["a"], random_path["a"])

    random_path3 = VectorDataset(df, copy=True)
    assert random_path3 == random_path
    assert random_path3["a"].base is None
    assert not np.may_share_memory(random_path3["a"], random_path["a"])

    # With copying, the resulting array is a honest copy of the original array
    df = random_path.to_dataframe(copy=True)
    a = df["a"].to_numpy()
    assert a.base is not random_path["a"]
    assert not np.may_share_memory(random_path3["a"], random_path["a"])

    data = random_path.data
    assert isinstance(data, dict)
    assert len(data) == 3

    # Here, when we grab the original data, the base is None
    random_path4 = VectorDataset(data, copy=False)
    assert random_path4 == random_path
    assert random_path4["a"].base is None
    assert np.may_share_memory(random_path4["a"], random_path["a"])

    random_path5 = VectorDataset(data, copy=True)
    assert random_path5 == random_path
    assert random_path5["a"].base is None
    assert not np.may_share_memory(random_path5["a"], random_path["a"])


def test_bad_time(random_geo_path: VectorDataset):
    """Test that bad time data is handled correctly."""
    data = random_geo_path.copy().data

    # time is successfully coerced to datetime64 (but values are nanoseconds after 1970)
    data.update(time=[1, 2, 3, 4])
    with pytest.warns(UserWarning, match="time"):
        GeoVectorDataset(data)

    # time cannot be converted
    data.update(time=["hello", "world", "con", "trail"])
    with pytest.warns(UserWarning, match="time"):
        with pytest.raises(ValueError, match="time"):
            GeoVectorDataset(data)

    data.update(
        time=[
            "2021-10-05 00:00:00Z",
            "2021-10-05 01:00:00Z",
            "2021-10-05 02:00:00Z",
            "2021-10-05 03:00:00Z",
        ]
    )
    df = pd.DataFrame(data)
    df["time"] = pd.to_datetime(df["time"])

    assert df["time"].dt.tz is not None
    with pytest.raises(ValueError, match="Column `time` must be timezone naive"):
        GeoVectorDataset(df)

    df["time"] = df["time"].dt.tz_localize(None)
    vector = GeoVectorDataset(df)
    assert vector["time"].dtype == np.dtype("datetime64[ns]")


def test_vector_from_array_like():
    """Confirm a Vector instance can be instanted from data castable to array."""
    data = {
        "a": [0, 1 / 3, 2 / 3],
        "b": [0, 0.4, 0.8],
    }
    vector = VectorDataset(data)
    assert vector.size == 3

    for k in vector:
        assert isinstance(vector[k], np.ndarray)


def test_vector_set_item_array_like():
    """Confirm __setitem__ correctly casts lists to arrays."""
    data = {
        "a": np.linspace(0, 1, 3),
        "b": np.arange(0, 1, 0.4),
    }
    vector = VectorDataset(data)
    assert vector.size == 3

    for k in vector:
        assert isinstance(vector[k], np.ndarray)

    # Correctly casts list
    c = [3, 2, 1]
    vector["c"] = c
    assert isinstance(vector["c"], np.ndarray)
    np.testing.assert_array_equal(vector["c"], c)
    assert vector["c"] is not c

    # If working with a numpy array, it's not copied
    d = np.array([3.3, 4.4, 5.5])
    vector["d"] = d
    assert vector["d"] is d



"""test utils module."""

from __future__ import annotations

import json
import pathlib

import numpy as np
import pandas as pd
import pytest
import xarray as xr

try:
    from pycontrails.utils.synthetic_flight import SAMPLE_AIRCRAFT_TYPES, SyntheticFlight
except ImportError:
    synthetic_flight_unavailable = True
    SAMPLE_AIRCRAFT_TYPES = []
else:
    synthetic_flight_unavailable = False


from pycontrails import Flight, MetDataset
from pycontrails.utils import types
from pycontrails.utils.iteration import chunk_list
from pycontrails.utils.json import NumpyEncoder
from pycontrails.utils.temp import remove_tempfile, temp_file, temp_filename
from pycontrails.utils.types import type_guard
from tests import BADA3_PATH, BADA4_PATH, BADA_AVAILABLE


def test_temp_files() -> None:
    """Test temp file handling."""
    temp_filename_ = temp_filename()
    assert isinstance(temp_filename_, str)
    assert not pathlib.Path(temp_filename_).is_file()

    with open(temp_filename_, "w") as f:
        f.write("test")

    # file persists after closing
    assert pathlib.Path(temp_filename_).is_file()

    # remove tempfile
    remove_tempfile(temp_filename_)
    assert not pathlib.Path(temp_filename_).is_file()


def test_temp_file_no_exception() -> None:
    """Check that `temp_file` automatically removes the temp file."""
    with temp_file() as tempfile:
        assert not pathlib.Path(tempfile).is_file()
        with open(tempfile, "w") as f:
            f.write("test")
        assert pathlib.Path(tempfile).is_file()

    # automatically removes file
    assert not pathlib.Path(tempfile).is_file()


@pytest.mark.parametrize("exc", [ValueError, RuntimeError, KeyError, IndexError, KeyboardInterrupt])
def test_temp_file_with_exception(exc: Exception) -> None:
    """Check that `temp_file` removes the temp file even when an exception is raised."""
    try:
        with temp_file() as tempfile:
            assert not pathlib.Path(tempfile).is_file()
            with open(tempfile, "w") as f:
                f.write("test")
            assert pathlib.Path(tempfile).is_file()
            raise exc
    except exc:  # type: ignore[misc]
        # automatically removes file
        assert not pathlib.Path(tempfile).is_file()


def test_numpy_encoder() -> None:
    """Test NumpyEncoder JSON utility."""
    int32 = 1
    float64 = 2.1
    array = [4, 5, 6]
    bool_ = True
    date_range = pd.date_range("2019-01-01 00:00:00", "2019-01-01 02:00:00", periods=10)
    _dict = {
        "int32": np.int32(int32),
        "float64": np.float64(float64),
        "array": np.array(array),
        "bool_": np.bool_(bool_),
        "series": pd.Series(array),
        "date_range": date_range,
    }

    jsonstr = json.dumps(_dict, cls=NumpyEncoder)
    loaded_dict = json.loads(jsonstr)

    assert loaded_dict["int32"] == int32
    assert loaded_dict["float64"] == float64
    assert loaded_dict["array"] == array
    assert loaded_dict["bool_"] == bool_
    assert loaded_dict["series"] == array
    assert loaded_dict["date_range"] == date_range.to_numpy().tolist()


@pytest.mark.skipif(synthetic_flight_unavailable, reason="synthetic_flight unavailable")
def test_synthetic_flight(met_cocip1: MetDataset) -> None:
    """Test SyntheticFlight class."""
    # synthetic flights
    bounds = {
        "longitude": np.array([-40, 40]),
        "latitude": np.array([-75, 75]),
        "level": np.array([175, 350]),
        "time": np.array([np.datetime64("2019-01-15"), np.datetime64("2019-02-16")]),
    }

    # no met, bada
    fl_gen = SyntheticFlight(aircraft_type="B737", bounds=bounds, seed=1, speed_m_per_s=240)
    fl = fl_gen()

    assert isinstance(fl, Flight)
    assert fl.attrs["aircraft_type"] == "B737"

    # random aircraft_type
    fl_gen = SyntheticFlight(bounds=bounds, seed=1, speed_m_per_s=240)
    fl = fl_gen()
    assert fl.attrs["aircraft_type"] in SAMPLE_AIRCRAFT_TYPES

    # with met
    fl_gen = SyntheticFlight(
        aircraft_type="B737",
        bounds=met_cocip1.coords,
        seed=1,
        speed_m_per_s=240,
        u_wind=met_cocip1["eastward_wind"],
        v_wind=met_cocip1["northward_wind"],
    )
    fl = fl_gen()

    assert isinstance(fl, Flight)


@pytest.mark.skipif(synthetic_flight_unavailable, reason="synthetic_flight unavailable")
@pytest.mark.skipif(not BADA_AVAILABLE, reason="No BADA or EDB Files")
@pytest.mark.parametrize("aircraft_type", SAMPLE_AIRCRAFT_TYPES)
def test_synthetic_flight_bada(aircraft_type: str) -> None:
    """Test BADA support in SyntheticFlight."""
    # synthetic flights
    bounds = {
        "longitude": np.array([-40, 40]),
        "latitude": np.array([-75, 75]),
        "level": np.array([175, 350]),
        "time": np.array([np.datetime64("2019-01-15"), np.datetime64("2019-02-16")]),
    }

    fl_gen = SyntheticFlight(
        aircraft_type=aircraft_type,
        bada3_path=BADA3_PATH,
        bada4_path=BADA4_PATH,
        bounds=bounds,
        seed=1,
    )
    fl = fl_gen()

    assert isinstance(fl, Flight)
    assert fl.attrs["aircraft_type"] == aircraft_type
    tas = fl.segment_true_airspeed()
    assert np.isnan(tas[-1])
    assert np.all(tas[:-1] > 100)
    assert np.all(tas[:-1] < 300)


def test_nan_mask() -> None:
    """Check that `apply_nan_mask_to_arraylike` returns correct type."""
    np_arr = np.arange(10).astype(float)
    nan_mask = np.zeros_like(np_arr).astype(bool)
    nan_mask[5] = True

    np_out = types.apply_nan_mask_to_arraylike(np_arr, nan_mask)
    assert isinstance(np_out, np.ndarray)
    assert np.isnan(np_out).nonzero()[0].item() == 5

    pd_arr = pd.Series(np_arr)
    pd_out = types.apply_nan_mask_to_arraylike(pd_arr, nan_mask)
    assert isinstance(pd_out, pd.Series)
    assert (pd_out.isna() == nan_mask).all()

    xr_arr = xr.DataArray(np_arr)
    nan_mask = xr.DataArray(nan_mask)
    xr_out = types.apply_nan_mask_to_arraylike(xr_arr, nan_mask)
    assert isinstance(xr_out, xr.DataArray)
    assert np.all(np.isnan(xr_out) == nan_mask)


def test_support_arraylike() -> None:
    """Test `support_arraylike` decorator."""
    np_arr = np.arange(10)

    @types.support_arraylike
    def func(arr):
        """Use a piecewise function here to mimic in internal use-cases."""
        condlist = [arr < 5]
        funclist = [lambda x: x**2, lambda x: 3 * x]
        return np.piecewise(arr, condlist, funclist)

    expected = [0, 1, 4, 9, 16, 15, 18, 21, 24, 27]

    # numpy case
    np_out = func(np_arr)
    assert isinstance(np_out, np.ndarray)
    np.testing.assert_array_equal(np_out, expected)

    # pandas case
    pd_arr = pd.Series(np_arr)
    pd_out = func(pd_arr)
    assert isinstance(pd_out, pd.Series)
    assert (pd_out == expected).all()

    # xarray case
    xr_arr = xr.DataArray(np_arr)
    xr_out = func(xr_arr)
    assert isinstance(xr_out, xr.DataArray)
    assert (xr_out == expected).all()

    # float, int case
    for i, e in enumerate(expected):
        assert func(i) == func(float(i)) == e


def test_chunk_list() -> None:
    """Test chunk list utility."""

    a = list(range(0, 50))
    for chunk in chunk_list(a, 10):
        assert len(chunk) == 10
        assert chunk[0] % 10 == 0


def test_type_guard(met_cocip1: MetDataset) -> None:
    """Test type_guard."""

    type_guard("str", str)
    type_guard(0, int)
    type_guard(met_cocip1, MetDataset)
    type_guard("str", (str, int))

    # returns object
    a = "test"
    b = type_guard(a, (str, int))
    assert a is b and isinstance(b, str)

    with pytest.raises(ValueError, match="must be of type"):
        type_guard("str", int)

    with pytest.raises(ValueError, match="pycontrails.core.met.MetDataset"):
        type_guard("str", (MetDataset, float))

    # custom error message
    with pytest.raises(ValueError, match="custom error"):
        type_guard("str", int, error_message="custom error")



"""ECMWF Data Access."""

from __future__ import annotations

from pycontrails.datalib.ecmwf.era5 import ERA5
from pycontrails.datalib.ecmwf.hres import HRES
from pycontrails.datalib.ecmwf.ifs import IFS
from pycontrails.datalib.ecmwf.variables import (
    ECMWF_VARIABLES,
    PRESSURE_LEVEL_VARIABLES,
    SURFACE_VARIABLES,
    CloudAreaFraction,
    CloudAreaFractionInLayer,
    PotentialVorticity,
    RelativeHumidity,
    RelativeVorticity,
    SpecificCloudIceWaterContent,
    SpecificCloudLiquidWaterContent,
    SurfaceSolarDownwardRadiation,
    TOAIncidentSolarRadiation,
    TopNetSolarRadiation,
    TopNetThermalRadiation,
)

__all__ = [
    "ERA5",
    "HRES",
    "IFS",
    "CloudAreaFractionInLayer",
    "SpecificCloudIceWaterContent",
    "TopNetSolarRadiation",
    "TopNetThermalRadiation",
    "RelativeHumidity",
    "RelativeVorticity",
    "SpecificCloudLiquidWaterContent",
    "TOAIncidentSolarRadiation",
    "CloudAreaFraction",
    "ECMWF_VARIABLES",
    "PRESSURE_LEVEL_VARIABLES",
    "SURFACE_VARIABLES",
]



"""Datalib utilities."""

from __future__ import annotations

import abc
import dataclasses
import hashlib
import logging
import pathlib
from datetime import datetime
from typing import Any, Sequence, Union

import numpy as np
import pandas as pd
import xarray as xr

from pycontrails.core import cache
from pycontrails.core.met import MetDataset, MetVariable
from pycontrails.utils.types import DatetimeLike

logger = logging.getLogger(__name__)

TimeInput = Union[str, DatetimeLike, Sequence[Union[str, DatetimeLike]]]
VariableInput = Union[
    str, int, MetVariable, np.ndarray, Sequence[Union[str, int, MetVariable, Sequence[MetVariable]]]
]
PressureLevelInput = Union[int, float, np.ndarray, Sequence[Union[int, float]]]

#: NetCDF engine to use for parsing netcdf files
NETCDF_ENGINE: str = "netcdf4"

#: Default chunking strategy when opening datasets with xarray
DEFAULT_CHUNKS: dict[str, int] = {"time": 1}


def parse_timesteps(time: TimeInput | None, freq: str | None = "1H") -> list[datetime]:
    """Parse time input into set of time steps.

    If input time is length 2, this creates a range of equally spaced time
    points between ``[start, end]`` with interval ``freq``.

    Parameters
    ----------
    time : TimeInput | None
        Input datetime(s) specifying the time or time range of the data [start, end].
        Either a single datetime-like or tuple of datetime-like with the first value
        the start of the date range and second value the end of the time range.
        Input values can be any type compatible with :meth:`pandas.to_datetime`.
    freq : str | None, optional
        Timestep interval in range.
        See https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#timeseries-offset-aliases
        for a list of frequency aliases.
        If None, returns input `time` as a list.
        Defaults to "1H".

    Returns
    -------
    list[datetime]
        List of unique datetimes.
        If input ``time`` is None, returns an empty list

    Raises
    ------
    ValueError
        Raises when the time has len > 2 or when time elements fail to be parsed with pd.to_datetime
    """  # noqa: E501

    if time is None:
        return []

    # confirm input is tuple or list-like of length 2
    if isinstance(time, (str, datetime, pd.Timestamp, np.datetime64)):
        time = (time, time)
    elif len(time) == 1:
        time = (time[0], time[0])
    elif len(time) != 2:
        raise ValueError("Input time bounds must have length < 2 and > 0")

    # convert all to pandas Timestamp
    try:
        timestamps = [pd.to_datetime(t) for t in time]
    except ValueError as e:
        raise ValueError(
            f"Failed to parse all time inputs with error {e}. Time input "
            "must be compatible with 'pd.to_datetime()'"
        )

    if freq is None:
        daterange = pd.DatetimeIndex([timestamps[0], timestamps[1]])
    else:
        # get date range that encompasses all whole hours
        daterange = pd.date_range(timestamps[0].floor(freq), timestamps[1].ceil(freq), freq=freq)

    # return list of datetimes
    return daterange.to_pydatetime().tolist()


def parse_pressure_levels(
    pressure_levels: PressureLevelInput, supported: list[int] | None = None
) -> list[int]:
    """Check input pressure levels are consistent type and ensure levels exist in ECMWF data source.

    Parameters
    ----------
    pressure_levels : PressureLevelInput
        Input pressure levels for data, in hPa (mbar)
        Set to [-1] to represent surface level.
    supported : list[int], optional
        List of supported pressures levels in data source

    Returns
    -------
    list[int]
        List of integer pressure levels supported by ECMWF data source

    Raises
    ------
    ValueError
        Raises ValueError if pressure level is not supported by ECMWF data source
    """
    # ensure pressure_levels is list-like
    if isinstance(pressure_levels, (int, float)):
        pressure_levels = [pressure_levels]

    # Cast array-like to list of ints
    # Use a new variable to satiate mypy
    pressure_levels_ = np.asarray(pressure_levels, dtype=int).tolist()

    # ensure pressure levels are valid
    for pl in pressure_levels_:
        if supported is not None and pl not in supported:
            raise ValueError(f"{pl} is not a valid pressure level {supported}")

    return pressure_levels_


def parse_variables(variables: VariableInput, supported: list[MetVariable]) -> list[MetVariable]:
    """Parse input variables.

    Parameters
    ----------
    variables : VariableInput
        Variable name, or sequence of variable names.
        i.e. ``"air_temperature"``, ``["air_temperature, relative_humidity"]``,
        ``[130]``, ``[AirTemperature]``, ``[[EastwardWind, NorthwardWind]]``
        If an element is a list of MetVariable, the first MetVariable that is
        supported will be chosen.
    supported : list[MetVariable]
        Supported MetVariable.

    Returns
    -------
    list[MetVariable]
        List of MetVariable

    Raises
    ------
    ValueError
        Raises ValueError if variable is not supported
    """
    parsed_variables: Sequence[str | int | MetVariable | Sequence[MetVariable]]
    met_var_list: list[MetVariable] = []

    # ensure input variables are a list of str
    if isinstance(variables, (str, int, MetVariable)):
        parsed_variables = [variables]
    elif isinstance(variables, np.ndarray):
        parsed_variables = variables.tolist()
    else:
        parsed_variables = variables

    # unpack list of supported str values from MetVariables
    short_names = [v.short_name for v in supported]
    standard_names = [v.standard_name for v in supported]
    long_names = [v.long_name for v in supported]

    # unpack list of support int values from Met Variables
    ecmwf_ids = [v.ecmwf_id for v in supported]
    grib1_ids = [v.grib1_id for v in supported]

    for var in parsed_variables:
        matched_variable: MetVariable | None = None

        if isinstance(var, MetVariable):
            if var in supported:
                matched_variable = var

        # list of MetVariable options
        # here we extract the first MetVariable in var that is supported
        elif isinstance(var, (list, tuple)):
            for v in var:
                # sanity check since we don't support other types as lists
                if not isinstance(v, MetVariable):
                    raise ValueError("Variable options must be of type MetVariable.")
                if v in supported:
                    matched_variable = v
                    break

        # int code
        elif isinstance(var, int):
            if var in ecmwf_ids:
                matched_variable = supported[ecmwf_ids.index(var)]
            elif var in grib1_ids:
                matched_variable = supported[grib1_ids.index(var)]

        # string reference
        elif isinstance(var, str):
            if var in short_names:
                matched_variable = supported[short_names.index(var)]
            elif var in standard_names:
                matched_variable = supported[standard_names.index(var)]
            elif var in long_names:
                matched_variable = supported[long_names.index(var)]

        if matched_variable is None:
            raise ValueError(
                f"{var} is not in supported parameters. "
                + f"Supported parameters include: {standard_names}"
            )

        # "replace" copies dataclass
        met_var_list.append(dataclasses.replace(matched_variable))

    return met_var_list


def parse_grid(grid: float, supported: list[float]) -> float:
    """Parse input grid spacing.

    Parameters
    ----------
    grid : float
        Input grid float
    supported : list[float]
        List of support grid values

    Returns
    -------
    float
        Parsed grid spacing

    Raises
    ------
    ValueError
        Raises ValueError when ``grid`` is not in supported
    """
    if grid not in supported:
        raise ValueError(f"Grid input must be one of {supported}")

    return grid


def round_hour(time: datetime, hour: int) -> datetime:
    """Round time to the nearest whole hour before input time.

    Parameters
    ----------
    time : datetime
        Input time
    hour : int
        Hour to round down time

    Returns
    -------
    datetime
        Rounded time

    Raises
    ------
    ValueError
        Description
    """
    if hour not in range(1, 24):
        raise ValueError("hour must be between [1, 23]")

    hour = np.floor(time.hour / hour) * hour
    return datetime(
        time.year,
        time.month,
        time.day,
        int(hour),
        0,
        0,
    )


class MetDataSource(abc.ABC):
    """Abstract class for wrapping meteorology data sources."""

    __slots__ = ("timesteps", "variables", "pressure_levels", "grid", "paths")

    #: List of individual timesteps from data source derived from :attr:`time`
    #: Use :func:`parse_time` to handle :class:`TimeInput`.
    timesteps: list[datetime]

    #: Variables requested from data source
    #: Use :func:`parse_variables` to handle :class:`VariableInput`.
    variables: list[MetVariable]

    #: List of pressure levels. Set to [-1] for data without level coordinate.
    #: Use :func:`parse_pressure_levels` to handle :class:`PressureLevelInput`.
    pressure_levels: list[int]

    #: Lat / Lon grid spacing
    grid: float | None

    #: Path to local source files to load.
    #: Set to the paths of files cached in :attr:`cachestore` if no
    #: ``paths`` input is provided on init.
    paths: str | list[str] | pathlib.Path | list[pathlib.Path] | None

    #: Cache store for intermediates while processing data source
    #: If None, cache is turned off.
    cachestore: cache.CacheStore | None

    def __repr__(self) -> str:
        _repr = (
            f"{self.__class__.__name__}\n\t"
            f"Timesteps: {[t.strftime('%Y-%m-%d %H') for t in self.timesteps]}\n\t"
            f"Variables: {self.variable_shortnames}\n\t"
            f"Pressure levels: {self.pressure_levels}\n\t"
            f"Grid: {self.grid}"
        )

        if self.paths is not None:
            _repr += f"\n\tPaths: {self.paths}"

        return _repr

    @abc.abstractmethod
    def __init__(
        self,
        time: TimeInput | None,
        variables: VariableInput,
        pressure_levels: PressureLevelInput = [-1],
        paths: str | list[str] | pathlib.Path | list[pathlib.Path] | None = None,
        grid: float | None = None,
        **kwargs: Any,
    ) -> None:
        ...

    @property
    def hash(self) -> str:
        """Generate a unique hash for this datasource.

        Returns
        -------
        str
            Unique hash for met instance (sha1)
        """
        hashstr = (
            f"{type(self).__name__}{self.timesteps}{self.variable_shortnames}{self.pressure_levels}"
        )
        return hashlib.sha1(bytes(hashstr, "utf-8")).hexdigest()

    @property
    def variable_shortnames(self) -> list[str]:
        """Return a list of variable short names.

        Returns
        -------
        list[str]
            Lst of variable short names.
        """
        return [v.short_name for v in self.variables]

    @property
    def variable_standardnames(self) -> list[str]:
        """Return a list of variable standard names.

        Returns
        -------
        list[str]
            Lst of variable standard names.
        """
        return [v.standard_name for v in self.variables]

    @property
    def pressure_level_variables(self) -> list[MetVariable]:
        """Parameters available from data source.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return []

    @property
    def single_level_variables(self) -> list[MetVariable]:
        """Parameters available from data source.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        return []

    @property
    def supported_variables(self) -> list[MetVariable]:
        """Parameters available from data source.

        Returns
        -------
        list[MetVariable] | None
            List of MetVariable available in datasource
        """
        if self.pressure_levels != [-1]:
            return self.pressure_level_variables

        return self.single_level_variables

    @property
    def supported_pressure_levels(self) -> list[int] | None:
        """Pressure levels available from datasource.

        Returns
        -------
        list[int] | None
            List of integer pressure levels for class.
            If None, no pressure level information available for class.
        """
        return None

    @property
    def _cachepaths(self) -> list[str]:
        """Return cache paths to local data files.

        Returns
        -------
        list[str]
            Path to local data files
        """
        return [self.create_cachepath(t) for t in self.timesteps]

    # -----------------------------
    # Abstract methods to implement
    # -----------------------------
    @abc.abstractmethod
    def download_dataset(self, times: list[datetime]) -> None:
        """Download data from data source for input times.

        Parameters
        ----------
        times : list[datetime]
            List of datetimes to download a store in cache
        """

    @abc.abstractmethod
    def create_cachepath(self, t: datetime) -> str:
        """Return cachepath to local data file based on datetime.

        Parameters
        ----------
        t : datetime
            Datetime of datafile

        Returns
        -------
        str
            Path to cached data file
        """

    @abc.abstractmethod
    def cache_dataset(self, dataset: xr.Dataset) -> None:
        """Cache data from data source.

        Parameters
        ----------
        dataset : xr.Dataset
            Dataset loaded from remote API or local files.
            The dataset must have the same format as the original data source API or files.
        """

    @abc.abstractmethod
    def open_metdataset(
        self,
        dataset: xr.Dataset | None = None,
        xr_kwargs: dict[str, int] | None = None,
        **kwargs: Any,
    ) -> MetDataset:
        """Open MetDataset from data source.

        This method should download / load any required datafiles and
        returns a MetDataset of the multi-file dataset opened by xarray.

        Parameters
        ----------
        dataset : xr.Dataset | None, optional
            Input :class:`xr.Dataset` loaded manually.
            The dataset must have the same format as the original data source API or files.
        xr_kwargs : dict[str, int] | None, optional
            Dictionary of keyword arguments passed into :func:`xarray.open_mfdataset`
            when opening files. Examples include "chunks", "engine", "parallel", etc.
            Ignored if ``dataset`` is input.
        **kwargs : Any
            Keyword arguments passed through directly into :class:`MetDataset` constructor.

        Returns
        -------
        MetDataset
            Meteorology dataset

        See Also
        --------
        :func:`xarray.open_mfdataset`
        """

    # ----------------------
    # Common utility methods
    # ----------------------
    def download(self, **xr_kwargs: Any) -> None:
        """Confirm all data files are downloaded and available locally in the :attr:`cachestore`.

        Parameters
        ----------
        **xr_kwargs
            Passed into :func:`xarray.open_dataset` via :meth:`is_datafile_cached`.
        """
        if times_to_download := self.list_timesteps_not_cached(**xr_kwargs):
            logger.debug(
                f"Not all files found in cachestore. Downloading times {times_to_download}"
            )
            self.download_dataset(times_to_download)
        else:
            logger.debug("All data files already in cache store")

    def list_timesteps_cached(self, **xr_kwargs: Any) -> list[datetime]:
        """Get a list of data files available locally in the :attr:`cachestore`.

        Parameters
        ----------
        **xr_kwargs
            Passed into :func:`xarray.open_dataset` via :meth:`is_datafile_cached`.
        """
        return [t for t in self.timesteps if self.is_datafile_cached(t, **xr_kwargs)]

    def list_timesteps_not_cached(self, **xr_kwargs: Any) -> list[datetime]:
        """Get a list of data files not available locally in the :attr:`cachestore`.

        Parameters
        ----------
        **xr_kwargs
            Passed into :func:`xarray.open_dataset` via :meth:`is_datafile_cached`.
        """
        return [t for t in self.timesteps if not self.is_datafile_cached(t, **xr_kwargs)]

    def is_datafile_cached(self, t: datetime, **xr_kwargs: Any) -> bool:
        """Check datafile defined by datetime for variables and pressure levels in class.

        If using a cloud cache store (i.e. :class:`cache.GCPCacheStore`) this is where the datafile
        will be mirrored to a local file for access.

        Parameters
        ----------
        t : datetime
            Datetime of datafile
        **xr_kwargs : Any
            Additional kwargs passed directly to :func:`xarray.open_mfdataset` when
            opening files. By default, the following values are used if not specified:

                - chunks: {"time": 1}
                - engine: "netcdf4"
                - parallel: True

        Returns
        -------
        bool
            True if data file exists for datetime with all variables and pressure levels,
            False otherwise
        """

        # return false if the cache is turned off
        if self.cachestore is None:
            return False

        # see if cache data file exists, and if so, get the file + path
        cache_path = self.create_cachepath(t)
        if self.cachestore.exists(cache_path):
            logger.debug(f"Cachepath {cache_path} exists, getting from cache.")
            # If GCP cache is used, this will download file and return the local mirrored path
            # If the local file already exists, this will return the local path
            disk_path = self.cachestore.get(cache_path)

            # check if all variables and pressure levels are in that path
            try:
                with self.open_dataset(disk_path, **xr_kwargs) as ds:
                    for var in self.variable_shortnames:
                        if var not in ds:
                            logger.warning(
                                f"Variable {var} not in downloaded dataset. "
                                f"Found variables: {ds.data_vars}"
                            )
                            return False

                    for pl in self.pressure_levels:
                        if pl not in ds["level"].values:
                            logger.warning(
                                f"Pressure Level {pl} not in downloaded dataset. "
                                f"Found pressure levels: {ds['level'].values}"
                            )
                            return False

                    logger.debug(f"All variables and pressure levels found in {cache_path}")
                    return True

            except OSError as err:
                if isinstance(self.cachestore, cache.GCPCacheStore):
                    # If a GCPCacheStore is used, remove the corrupt file and
                    # try again.
                    # If the file is corrupt in the GCP bucket, we'll
                    # get stuck in an infinite loop here
                    logger.warning(
                        "Found corrupt file %s on local disk. Try again to download from %s.",
                        disk_path,
                        self.cachestore,
                    )
                    self.cachestore.clear_disk(disk_path)
                    return self.is_datafile_cached(t, **xr_kwargs)

                raise OSError(
                    f"Unable to open NETCDF file at {disk_path} "
                    "This may be due to a incomplete download. "
                    f"Consider manually removing {disk_path} and retrying."
                ) from err

        logger.debug(f"Cachepath {cache_path} does not exist in cache")
        return False

    def open_dataset(
        self,
        disk_paths: str | list[str] | pathlib.Path | list[pathlib.Path],
        **xr_kwargs: Any,
    ) -> xr.Dataset:
        """Open multi-file dataset in xarray.

        Parameters
        ----------
        disk_paths : str | list[str] | pathlib.Path | list[pathlib.Path]
            list of string paths to local files to open
        **xr_kwargs : Any
            Additional kwargs passed directly to :func:`xarray.open_mfdataset` when
            opening files. By default, the following values are used if not specified:

                - chunks: {"time": 1}
                - engine: "netcdf4"
                - parallel: True

        Returns
        -------
        xr.Dataset
            Open xarray dataset
        """
        xr_kwargs.setdefault("engine", NETCDF_ENGINE)
        xr_kwargs.setdefault("chunks", DEFAULT_CHUNKS)
        xr_kwargs.setdefault("parallel", True)
        return xr.open_mfdataset(disk_paths, **xr_kwargs)



"""Utilities for :class:`Cocip` and :class:`CocipGrid`."""

from __future__ import annotations

import dataclasses
import logging
import warnings
from typing import TYPE_CHECKING, Any

import numpy as np
import numpy.typing as npt
import pandas as pd

from pycontrails.core import coordinates
from pycontrails.core.met import MetDataset
from pycontrails.core.vector import GeoVectorDataset

if TYPE_CHECKING:
    import tqdm

logger = logging.getLogger(__name__)

# Crude constants used to estimate model runtime
# If we seek more accuracy, these should be seasonally adjusted
# The first constant is the proportion of points generating persistent contrails
# that survive to the end of the hour
# The second constants it the probability of a contrail surviving an hour of
# ongoing model evolution
HEURISTIC_INITIAL_SURVIVAL_RATE = 0.1
HEURISTIC_EVOLUTION_SURVIVAL_RATE = 0.8


class CocipTimeHandlingMixin:
    """Support :class:`Cocip` and :class:`CocipGrid` time handling."""

    params: dict[str, Any]
    source: GeoVectorDataset | MetDataset
    met: MetDataset
    rad: MetDataset

    #: Convenience container to hold time filters
    #: See :meth:`CocipGrid.eval` for usage.
    timedict: dict[np.datetime64, np.ndarray]

    def validate_time_params(self) -> None:
        """Raise a `ValueError` if `met_slice_dt` is not a multiple of `dt_integration`."""
        met_slice_dt: np.timedelta64 | None = self.params.get("met_slice_dt")

        if met_slice_dt is None:
            # TODO: Reset to something big enough to cover the entire met dataset
            raise NotImplementedError
        if met_slice_dt == np.timedelta64(0, "s"):
            raise ValueError("met_slice_dt must be a positive timedelta")

        dt_integration: np.timedelta64 = self.params["dt_integration"]
        ratio = met_slice_dt / dt_integration
        if not ratio.is_integer():
            raise ValueError(
                f"met_slice_dt ({met_slice_dt}) must be a multiple "
                f"of dt_integration ({dt_integration})"
            )

        # I'm not sure how necessary this is ...
        met_time_diff = np.diff(self.met.data["time"].values)
        ratios = met_slice_dt / met_time_diff
        if np.any(ratios != ratios.astype(int)):
            raise ValueError(
                f"met_slice_dt ({met_slice_dt}) must be a multiple "
                "of the time difference between met time steps "
                f"({met_time_diff})"
            )

        # Validate met_slice_dt
        met_time_diff = np.diff(self.met.data["time"].values).astype("timedelta64[h]")
        if np.unique(met_time_diff).size > 1:
            raise NotImplementedError("CocipGrid only supports met with constant time diff.")
        met_time_res = met_time_diff[0]

        n_hours = met_slice_dt / met_time_res
        if n_hours < 1 or not n_hours.is_integer():
            raise ValueError(
                f"Parameter `met_slice_dt` must be a positive multiple of {met_time_res}."
            )

    @property
    def source_time(self) -> npt.NDArray[np.datetime64]:
        """Return the time array of the :attr:`source` data."""
        if not hasattr(self, "source"):
            raise AttributeError("source not set")
        if isinstance(self.source, GeoVectorDataset):
            return self.source["time"]
        if isinstance(self.source, MetDataset):
            return self.source.variables["time"].values
        raise TypeError(f"Cannot calculate timesteps for {self.source}")

    def attach_timedict(self) -> None:
        """Attach or update :attr:`timedict`.

        This attribute is a dictionary of the form::

            {t: filt}

        where the key ``t`` is the time of the start of the met slice and the value
        ``filt`` is a :class:`numpy.ndarray` of the same shape as the
        the time source. Specifically, ``filt`` is a boolean array that can
        be used to filter the time source.

        Presently, keys are hard-coded with :attr:`dtype` ``datetime64[h]``.

        The keys of this dictionary can be used to slice the ``met`` dataset
        in the ``time`` dimension. However, this must be done with care in case
        the met dataset is not aligned with the time source.
        """
        met_slice_dt = self.params.get("met_slice_dt")
        if met_slice_dt is None:
            raise ValueError("met_slice_dt must be set")
        self.validate_time_params()

        # Cast to pandas to use ceil and floor methods below
        met_slice_dt = pd.to_timedelta(met_slice_dt)
        source_time = self.source_time
        tmin = pd.to_datetime(source_time.min())
        tmax = pd.to_datetime(source_time.max() + self.params["max_age"])
        self._check_met_rad_time(tmin, tmax)

        # Ideally we'd use the keys to index into the met dataset
        # with met.data.sel(time=slice(t1, t2)), but this is not even
        # possible the the rad data because of time shifting.
        # So, at the very least, we'll want to use np.searchsorted
        # to find the indices of the met dataset that correspond to
        # the time slice. See _load_met_slices
        t_start = tmin.floor(met_slice_dt)
        t_end = tmax.ceil(met_slice_dt)

        met_times = np.arange(t_start, t_end + met_slice_dt, met_slice_dt).astype("datetime64[h]")
        zipped = zip(met_times, met_times[1:])
        self.timedict = {t1: (source_time >= t1) & (source_time < t2) for t1, t2 in zipped}

    def _check_met_rad_time(self, tmin: pd.Timestamp, tmax: pd.Timestamp) -> None:
        if self.met.data["time"].min() > tmin or self.met.data["time"].max() < tmax:
            warnings.warn(
                "Parameter 'met' is too short in the time dimension. "
                "Include additional time in 'met' or reduce 'max_age' parameter. "
                f"Model start time: {tmin} Model end time: {tmax}"
            )
        if self.rad.data["time"].min() > tmin or self.rad.data["time"].max() < tmax:
            warnings.warn(
                "Parameter 'rad' is too short in the time dimension. "
                "Include additional time in 'rad' or reduce 'max_age' parameter."
                f"Model start time: {tmin} Model end time: {tmax}"
            )

    def init_pbar(self) -> "tqdm.tqdm" | None:
        """Initialize a progress bar for model evaluation."""

        if not self.params["show_progress"]:
            return None

        try:
            from tqdm.auto import tqdm
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                f"Running model {type(self).__name__}  with parameter "
                "show_progress=True requires the 'tqdm' module, which can be "
                "installed with 'pip install tqdm'. "
                "Alternatively, set model parameter 'show_progress=False'."
            ) from e

        estimate = self._estimate_runtime()

        # We call update on contrail initialization and at each evolution step
        total = sum(e.n_steps_new_vectors + e.n_steps_old_vectors for e in estimate.values())
        # We also call update on each met loading step
        total += len(estimate)

        return tqdm(total=total, desc=f"{type(self).__name__} eval")

    def _estimate_runtime(self) -> dict[np.datetime64, CocipRuntimeStats]:
        """Calculate number of new meshes and predict number of persistent meshes by met slice.

        Returns
        -------
        dict[np.datetime64, CocipRuntimeStats]
            Estimate of the runtime for each met slice.
        """

        met_slice_dt = self.params["met_slice_dt"]
        if met_slice_dt is None:
            raise ValueError("met_slice_dt must be set")
        if not hasattr(self, "timedict"):
            raise AttributeError("First call attach_met_slice_timedict")

        dt_integration = self.params["dt_integration"]
        met_slice_dt = self.params["met_slice_dt"]
        if isinstance(self.source, MetDataset):
            n_splits = self._grid_spatial_n_splits()
        else:
            split_size = (
                self.params["target_split_size_pre_SAC_boost"] * self.params["target_split_size"]
            )

        # Adjust scaling factors when met_slice_dt is more than a single hour
        met_slice_scale = met_slice_dt / np.timedelta64(1, "h")

        # Mirror the structure of timedict to estimate the runtime
        estimate: dict[np.datetime64, CocipRuntimeStats] = {}
        for t_start, filt in self.timedict.items():
            t_next = t_start + met_slice_dt
            t_prev = t_start - met_slice_dt

            times_in_filt = self.source_time[filt]
            n_steps_by_time = np.ceil((t_next - times_in_filt) / dt_integration)

            # Tricky logic for different sources
            if isinstance(self.source, MetDataset):
                n_new_vectors = len(times_in_filt) * n_splits
                n_steps_new_vectors = int(n_steps_by_time.sum() + 1) * n_splits
            else:
                if times_in_filt.size:
                    n_new_vectors = max(int(times_in_filt.size / split_size), 1)
                    n_steps_new_vectors = int(n_steps_by_time.max() + 1) * n_new_vectors
                else:
                    n_new_vectors = 0
                    n_steps_new_vectors = 0

            prev_stats = estimate.get(t_prev)
            if prev_stats is None:
                prev_stats = CocipRuntimeStats(t_prev, 0, 0, 0, 0)

            n_old_vectors_float = int(
                HEURISTIC_INITIAL_SURVIVAL_RATE * prev_stats.n_new_vectors
                + HEURISTIC_EVOLUTION_SURVIVAL_RATE**met_slice_scale * prev_stats.n_old_vectors
            )
            n_old_vectors = max(round(n_old_vectors_float), 1)
            n_steps_old_vectors = n_old_vectors * (met_slice_dt // dt_integration).item()
            estimate[t_start] = CocipRuntimeStats(
                t_start=t_start,
                n_new_vectors=n_new_vectors,
                n_steps_new_vectors=n_steps_new_vectors,
                n_old_vectors=n_old_vectors,
                n_steps_old_vectors=n_steps_old_vectors,
            )

        return estimate

    def _grid_spatial_n_splits(self) -> int:
        """Compute the number of vector "spatial" splits at a single time.

        Helper method used in :meth:`_estimate_runtime` and :meth:`_generate_new_grid_vectors`.

        This method assumes :attr:`source` is a :class:`MetDataset`.

        Returns
        -------
        int
            The number of spatial splits.
        """
        grid_size = (
            self.source.data["longitude"].size
            * self.source.data["latitude"].size
            * self.source.data["level"].size
        )
        split_size = int(
            self.params["target_split_size_pre_SAC_boost"] * self.params["target_split_size"]
        )
        return max(grid_size // split_size, 1)

    def _load_met_slices(
        self, start: np.datetime64, pbar: "tqdm.tqdm" | None = None
    ) -> tuple[MetDataset, MetDataset]:
        """Load met and rad slices for interpolation.

        :attr:`met` and :attr:`rad` are sliced by `slice(start, start + .met_slice_dt)`

        Parameters
        ----------
        start : np.datetime64
            Start of time domain of interest. Does not need to have hour resolution,
            but we will get runtime errors in other methods if not.
        pbar : tqdm.tqdm | None, optional
            Progress bar. The :meth:`pbar.update` method is called after both slices
            are loaded.

        Returns
        -------
        met : MetDataset
            Met data sliced to the time domain of interest.
        rad : MetDataset
            Rad data sliced to the time domain of interest.

        Raises
        ------
        NotImplementedError
            If :attr:`met` data has finer than hourly "time" resolution
        RuntimeError
            If ``met`` or ``rad`` time slices do not not contain at least two time values.
        """

        # Only support met with hourly timestamps
        time = self.met.variables["time"].values
        remainder = time - time.astype("datetime64[h]")
        if np.any(remainder.astype(float)):
            raise NotImplementedError("Only support met data with hourly time coordinates.")

        request = start, start + self.params["met_slice_dt"]
        buffer = np.timedelta64(0, "h"), np.timedelta64(0, "h")
        met_sl = coordinates.slice_domain(time, request, buffer)
        rad_sl = coordinates.slice_domain(self.rad.variables["time"].values, request, buffer)

        logger.debug("Update met slices. Start: %s, Stop: %s", met_sl.start, met_sl.stop)
        logger.debug("Update rad slices. Start: %s, Stop: %s", rad_sl.start, rad_sl.stop)
        xr_met_slice = self.met.data.isel(time=met_sl)
        xr_rad_slice = self.rad.data.isel(time=rad_sl)

        # We no longer require time two slices for linear interpolation, but for the
        # sake of contrail evolution, we expect to always use two. See method
        # _evolve_vector, where we explicitly unpack the time domain into two variables.
        if len(xr_met_slice["time"]) < 2:
            raise RuntimeError("Malformed met slice.")
        if len(xr_rad_slice["time"]) < 2:
            raise RuntimeError("Malformed rad slice.")

        # If data is already loaded into memory, calling load will not waste memory
        met_slice = MetDataset(xr_met_slice, copy=False)
        rad_slice = MetDataset(xr_rad_slice, copy=False)

        if pbar is not None:
            pbar.update()
        return met_slice, rad_slice


@dataclasses.dataclass
class CocipRuntimeStats:
    """Support for estimating runtime and progress bar."""

    t_start: np.datetime64
    n_new_vectors: int
    n_old_vectors: int
    n_steps_new_vectors: int
    n_steps_old_vectors: int
